import sys
import os
import logging
from logging import INFO
from time import localtime, strftime
from typing import Dict

class Logger(logging.getLoggerClass()):
    """Logger class used to manage the log routing and writes logs on different files."""

    def __init__(self, format: str = "%(asctime)s | %(levelname)s | %(message)s", level: int = INFO):
        """Logger initialization, sets default logging format and log level 

        Parameters
        ----------
        format : str, optional
            Log message format, by default "%(asctime)s | %(levelname)s | %(message)s"
        level : int, optional
            Log level, by default INFO
        """

        self.logger = logging.getLogger("logger")
        self.important_logger = logging.getLogger("important_logger")

        # Initial construct.
        self.format = format
        self.level = level
        self.name = None
        self.log_path = None

    def initialize_logger(self, log_path : str, name: str):
        """Initialize the logging session by creating the logging files 
        <name>_<YYYYMMDD>.log and <name>_<YYYYMMDD>_verbose.log . If a new name 
        is passed then a new log session and files are created aswell

        Parameters
        ----------
        log_path : str
            Path where Logging will save the logs
        name : str
            Logging session name
        """
        self.name = name
        self.logging_folder = log_path

        # Logger configuration.
        self.console_formatter = logging.Formatter(self.format)
        self.console_logger = logging.StreamHandler(sys.stdout)
        self.console_logger.setFormatter(self.console_formatter)

        # Complete logging config.
        self.logger.setLevel(self.level)
        self.important_logger.setLevel(self.level)

        # Logger configuration.
        self.file_formatter = logging.Formatter(self.format)

        # check if folder exists
        if not os.path.exists(self.logging_folder):
            os.makedirs(self.logging_folder)
        self.file_logger = logging.FileHandler(os.path.join(
            self.logging_folder, self.name + "_verbose_" + strftime("%Y%m%d", localtime()) + ".log"))
        self.file_logger.setFormatter(self.file_formatter)
        self.important_file_logger = logging.FileHandler(
            os.path.join(self.logging_folder, self.name + "_" + strftime("%Y%m%d", localtime()) + ".log"))
        self.important_file_logger.setFormatter(self.file_formatter)

        # Complete logging config.
        handler_list = []
        for handler in self.logger.handlers:
            try:
                handler_list.append(os.path.split(handler.baseFilename)[1])

            except AttributeError:
                pass
        
        if self.name + "_verbose_" + strftime("%Y%m%d", localtime()) + ".log" not in handler_list:
            self.logger.addHandler(self.console_logger)
            self.logger.addHandler(self.file_logger)
        if self.name + "_" + strftime("%Y%m%d", localtime()) + ".log" not in handler_list:
            self.important_logger.addHandler(self.important_file_logger)

    def info(self, msg: str, extra: Dict = None, important: bool = False):
        """Generates a INFO log level row

        Parameters
        ----------
        msg : str
            Message associated to the log line
        extra : Dict, optional
            Dictionary containing the missing FORMAT attributes 
            needed to generate the log message, by default None
        important : bool, optional
           Specify wether to write or not on main .log file , by default False
        """
        self.logger.info(msg, extra=extra)
        if important:
            self.important_logger.info(msg, extra=extra)

    def error(self, msg: str, extra: Dict = None):
        """Generates an ERROR log level row

        Parameters
        ----------
        msg : str
            Message associated to the log line
        extra : Dict, optional
            Dictionary containing the missing FORMAT attributes 
            needed to generate the log message, by default None
        """
        self.logger.error(msg, extra=extra)
        self.important_logger.error(msg, extra=extra)

    def exception(self, msg: str, extra: Dict = None):
        """Generates an EXCEPTION log level row

        Parameters
        ----------
        msg : str
            Message associated to the log line
        extra : Dict, optional
            Dictionary containing the missing FORMAT attributes 
            needed to generate the log message, by default None
        """
        self.logger.exception(msg, extra=extra)
        self.important_logger.exception(msg, extra=extra)

    def warning(self, msg: str, extra: Dict = None):
        """Generates a WARNING log level row

        Parameters
        ----------
        msg : str
            Message associated to the log line
        extra : Dict, optional
            Dictionary containing the missing FORMAT attributes 
            needed to generate the log message, by default None
        """
        self.logger.warning(msg, extra=extra)

    def debug(self, msg: str, extra: Dict = None):
        """Generates a DEBUG log level row

        Parameters
        ----------
        msg : str
            Message associated to the log line
        extra : Dict, optional
            Dictionary containing the missing FORMAT attributes 
            needed to generate the log message, by default None
        """
        self.logger.debug(msg, extra=extra)

    # def profile_memory_usage(self):
    #     memory_usage = dict(psutil.virtual_memory()._asdict())
    #     memory_usage_log = str(
    #         {key: round(el / (1024 ** 3), 3) if key != "percent" else el for key, el in memory_usage.items()})
    #     self.logger.info("Memory usage: " + memory_usage_log)


logger = Logger()



from typing import Any, Iterable
import pytest

from test_tc.utility.check_utils import check_not_isinstance, check_not_in_iterable
from test_tc.utility.exceptions import InputTypeError, ConversionError
from test_tc.utility.tele_logger import logger

## check_not_isinstance ##
def check_not_isinstance(obj, data_type, func):
    if not isinstance(obj, data_type):
        logger.error(f"Type input obj {type(obj)} must be {data_type}")
        raise InputTypeError(func)

@pytest.mark.parametrize(
    "obj, data_type, func",
    [
        (5, str, "function1"),
        ("string", int, "function2"),
        ([1, 2, 3], dict, "function3"),
    ]
)

def test_check_not_isinstance(obj, data_type, func):
    with pytest.raises(InputTypeError):
        check_not_isinstance(obj, data_type, func)

@pytest.mark.parametrize(
    "obj, data_type, func",
    [
        (5, int, "function1"),
        ("string", str, "function2"),
        ([1, 2, 3], list, "function3"),
    ]
)

def test_check_not_isinstance_OK(obj, data_type, func):
    check_not_isinstance(obj, data_type, func)

## check_not_in_iterable ##
def check_not_in_iterable(obj: Any, iterable: Iterable, func: str):
    if obj not in iterable:
        logger.error(f"Invalid value for {obj}. It must be in {iterable}")
        raise InputTypeError(func)

@pytest.mark.parametrize(
    "obj, iterable, func",
    [
        (5, [1, 2, 3], "function1"),
        ("string", ["abc", "def", "ghi"], "function2"),
        (False, [True, True, None], "function3"),
    ]
)
def test_check_not_in_iterable(obj, iterable, func):
    with pytest.raises(InputTypeError):
        check_not_in_iterable(obj, iterable, func)

@pytest.mark.parametrize(
    "obj, iterable, func",
    [
        (5, [1, 2, 5], "function1"),
        ("string", ["abc", "string", "ghi"], "function2"),
        (False, [True, False, None], "function3"),
    ]
)

def test_check_not_in_iterable_OK(obj, iterable, func):
    check_not_in_iterable(obj, iterable, func)

## check_datatype_convertible ##
def check_datatype_convertible(obj: Any, data_type_conversion: Any, func: str): 
    try:
        data_type_conversion(obj)
    except (ValueError, TypeError):
        raise ConversionError(func = func)

def test_check_datatype_convertible():
    # Test a valid conversion
    obj = "123"
    data_type_conversion = int
    func = "test_check_datatype_convertible"
    assert check_datatype_convertible(obj, data_type_conversion, func) == None

    # Test an invalid conversion
    obj = "abc"
    data_type_conversion = int
    func = "test_check_datatype_convertible"
    with pytest.raises(ConversionError):
        check_datatype_convertible(obj, data_type_conversion, func)


## test ProphetPrerocess class ##
# The class is composed by the methods tested in test_datapreparation_utils.py 
# So, we just check if the output of the preprocessing can be the input for a prophet model


import prophet
import pytest
import pandas as pd

from test_tc.datapreparation.model_preproc import ProphetPreprocess


## test get_hierarchical_df ##
def sample_data_hier():
    # Sample data for testing
    data = {
        "date_col": pd.date_range("2023-07-01", periods=5, freq="D"),
        "zone": ["North", "North", "South", "South", "West"],
        "city": ["Milan", "Turin", "Rome", "Naples", "Florence"]
    }
    return pd.DataFrame(data)

import warnings
def test_ProphetPreprocess_df():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        hierarchy = {
            "level1": "zone",
            "level2": "city"
        }
        time_granularity = "D"
        date_col = "date_col"
        missing_data_strategy = ''

        # Sample data for testing
        df = sample_data_hier()

        prophet_preproc = ProphetPreprocess(hierarchy, time_granularity, date_col, missing_data_strategy)

        X, S, tags = prophet_preproc.transform(df)
        
        # Check if X is ready to be used for Prophet
        Prophet = prophet.Prophet()
        Prophet.fit(X)








import pytest
import pandas as pd
from tabulate import tabulate
from test_tc.utility.exceptions import InputTypeError, ColumnNotFound, ConversionError

### InputTypeError ###
def test_input_type_error_message():
    func_name = "my_function"
    error_message = "Wrong input types have been given to the function."
    error = InputTypeError(func_name)
    assert str(error) == f"{func_name} encountered an error: {error_message}"

def test_input_type_error_custom_message():
    func_name = "my_function"
    custom_message = "Invalid input provided."
    error = InputTypeError(func_name, custom_message)
    assert str(error) == f"{func_name} encountered an error: {custom_message}"

def test_column_not_found_exception():
    # Sample data
    df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})

    # Check if the ColumnNotFound exception is raised
    with pytest.raises(ColumnNotFound) as exc_info:
        raise ColumnNotFound(func='my_function', column='C', data=df)

    # Verify the exception message
    expected_message = f"my_function encountered an error: The following column is not in the data columns list -> C\n\n{tabulate(df, headers='keys', tablefmt='psql')}"
    assert str(exc_info.value) == expected_message

def test_conversion_error_default_message():
    func_name = "test_function"
    with pytest.raises(ConversionError) as exc_info:
        raise ConversionError(func_name)

    assert str(exc_info.value) == f"{func_name} encountered an error: The object cannot be converted"

def test_conversion_error_custom_message():
    func_name = "test_function"
    custom_message = "This is a custom error message."
    with pytest.raises(ConversionError) as exc_info:
        raise ConversionError(func_name, custom_message)

    assert str(exc_info.value) == f"{func_name} encountered an error: {custom_message}"


import pytest
import pandas as pd
import numpy as np

from test_tc.algorithms.utils import * 


## test_train_val_test_split.py ## 
# Helper function to create a sample DataFrame
def create_sample_dataframe():
    data = {
        'unique_id': [1, 1, 1, 2, 2, 2, 3, 3, 3],
        'feature_1': [10, 20, 30, 40, 50, 60, 70, 80, 90],
        'feature_2': [100, 200, 300, 400, 500, 600, 700, 800, 900],
    }
    data = pd.DataFrame(data).set_index('unique_id')
    return data

# Test cases
def test_train_val_test_split():
    # Create a sample DataFrame
    X = create_sample_dataframe()

    # Define val_size and test_size
    val_size = 1
    test_size = 1

    # Call the function to get the splits
    train_df, valid_df, test_df = train_val_test_split(X, val_size, test_size)
    # Assertions to check if the returned objects are DataFrames
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(valid_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Assertions to check the shape of the returned DataFrames
    assert train_df.shape[0] == X.shape[0] - valid_df.shape[0] - test_df.shape[0]

    # Assertions to check if the unique_id index is set correctly
    assert train_df.index.name == 'unique_id'
    assert valid_df.index.name == 'unique_id'
    assert test_df.index.name == 'unique_id'

    # Assertions to check if all unique_ids in the validation and test sets are also present in the train set
    assert set(valid_df.index).union(set(test_df.index)).issubset(set(train_df.index))
  
## test literal_evaluation ##
def test_literal_evaluation():
    # Test case 1: Valid literals
    values_list = ['True', 'False', '42', '3.14', "'hello'", '[1, 2, 3]', '{"a": 1, "b": 2}']
    expected_result = [True, False, 42, 3.14, 'hello', [1, 2, 3], {"a": 1, "b": 2}]
    assert literal_evaluation(values_list) == expected_result

    # Test case 2: Invalid literals (should return the same values as strings)
    values_list = ['invalid_literal_1', 'invalid_literal_2', 'invalid_literal_3']
    expected_result = ['invalid_literal_1', 'invalid_literal_2', 'invalid_literal_3']
    assert literal_evaluation(values_list) == expected_result

    # Test case 3: Empty list
    values_list = []
    expected_result = []
    assert literal_evaluation(values_list) == expected_result

    # Test case 4: Mix of valid and invalid literals
    values_list = ['True', 'invalid_literal', '42', '[1, 2, 3]', 'invalid_literal', '{"a": 1, "b": 2}']
    expected_result = [True, 'invalid_literal', 42, [1, 2, 3], 'invalid_literal', {"a": 1, "b": 2}]
    assert literal_evaluation(values_list) == expected_result

    # Test case 5: Single element list
    values_list = ['True']
    expected_result = [True]
    assert literal_evaluation(values_list) == expected_result

    # Test case 6: List with a nested list
    values_list = ['[1, [2, 3], 4]']
    expected_result = [[1, [2, 3], 4]]
    assert literal_evaluation(values_list) == expected_result

# test_generate_values.py
def test_generate_values():
    # Test case 1: Valid input with specified step
    dictionary = {'min': 1, 'max': 10, 'step': 2}
    default_step = 1
    expected_result = [1, 3, 5, 7, 9, 10]
    assert generate_values(dictionary, default_step) == expected_result

    # Test case 2: Valid input with step set to -1 (min value used as step)
    dictionary = {'min': 5, 'max': 20, 'step': -1}
    default_step = 2
    expected_result = [5, 10, 15, 20]
    assert generate_values(dictionary, default_step) == expected_result

    # Test case 3: Valid input with step set to -1 (default step used as step since min = 0)
    dictionary = {'min': 0, 'max': 10, 'step': -1}
    default_step = 1.5
    expected_result = [0, 1.5, 3.0, 4.5, 6.0, 7.5, 9.0, 10]
    assert generate_values(dictionary, default_step) == expected_result

    # Test case 4: Valid input with step set to 'None' (default step used as step)
    dictionary = {'min': 100, 'max': 200, 'step': 'None'}
    default_step = 20
    expected_result = list(np.arange(100, 200 + default_step, default_step))
    assert generate_values(dictionary, default_step) == expected_result

    # Test case 5: Invalid input (missing 'step' key in dictionary)
    dictionary = {'min': 0, 'max': 1}
    default_step = 0.1
    with pytest.raises(Exception):
        generate_values(dictionary, default_step)

    # Test case 6: Invalid input (step greater than values range)
    dictionary = {'min': 1, 'max': 10, 'step': 15}
    default_step = 2
    with pytest.raises(ValueError):
        generate_values(dictionary, default_step)

    # Test case 7: Invalid input (max_value <= min_value)
    dictionary = {'min': 5, 'max': 3, 'step': 1}
    default_step = 1
    with pytest.raises(ValueError):
        generate_values(dictionary, default_step)

## test grid_values_hyperparameters ##
# the function wraps the ones above so we will test only its result
def get_config_hyperparameters():
    config_hyperparameters = {
        'changepoint_prior_scale' : {'min': 10, 'max' : 20, 'step' : 5},
        'seasonality_prior_scale' :  {'min': 10, 'max' : 20, 'step' : 5},
        'weekly_seasonality' : ['True'],
        'yearly_seasonality' : ['True']}
    return config_hyperparameters

def test_grid_values_hyperparameters():
    # Good configuration
    config_hyperparameters = get_config_hyperparameters()
    expected_result = {
        'changepoint_prior_scale' : [10, 15, 20],
        'seasonality_prior_scale' :  [10, 15, 20],
        'weekly_seasonality' : [True],
        'yearly_seasonality' : [True]}
    default_steps = {'changepoint_prior_scale' : 0.1,
                     'seasonality_prior_scale' : 1.0}
    assert grid_values_hyperparameters(config_hyperparameters, default_steps) == expected_result

## create_zero_dataframe ##
def test_create_zero_dataframe():
    # Test case 1: Check if the DataFrame is created with correct shape and zero values
    columns = ['A', 'B', 'C']
    n_rows = 5
    df = create_zero_dataframe(columns, n_rows)

    # Check DataFrame shape
    assert df.shape == (n_rows, len(columns))

    # Check if all values are zero
    assert (df.values == 0).all()

    # Test case 2: Check for an empty DataFrame
    columns = []
    n_rows = 0
    df = create_zero_dataframe(columns, n_rows)

    # Check if the DataFrame is empty
    assert df.empty

    # Test case 3: Check for a single-column DataFrame with zero rows
    columns = ['A']
    n_rows = 0
    df = create_zero_dataframe(columns, n_rows)

    # Check DataFrame shape and if the column is present
    assert df.shape == (n_rows, len(columns))
    assert 'A' in df.columns

## test write_dataframe_locally ##
def test_write_dataframes_locally(tmpdir):
    # Convert tmpdir to a str for the test case
    save_data_folder = str(tmpdir)

    # Test case 1: Write a single DataFrame
    df1 = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})
    file_name1 = 'file1.csv'
    to_save = [(df1, file_name1)]
    write_dataframes_locally(to_save, save_data_folder)

    # Check if the file was written
    file_path1 = os.path.join(save_data_folder, file_name1)
    assert os.path.exists(file_path1)
    assert os.path.isfile(file_path1)

    # Test case 2: Write multiple DataFrames
    df2 = pd.DataFrame({'C': [7, 8, 9], 'D': [10, 11, 12]})
    file_name2 = 'file2.csv'
    df3 = pd.DataFrame({'E': [13, 14, 15], 'F': [16, 17, 18]})
    file_name3 = 'file3.csv'
    to_save = [(df2, file_name2), (df3, file_name3)]
    write_dataframes_locally(to_save, save_data_folder)

    # Check if the files were written
    file_path2 = os.path.join(save_data_folder, file_name2)
    assert os.path.exists(file_path2)
    assert os.path.isfile(file_path2)

    file_path3 = os.path.join(save_data_folder, file_name3)
    assert os.path.exists(file_path3)
    assert os.path.isfile(file_path3)

## test write_json_locally ## 
def test_write_json_locally(tmpdir):
    # Convert tmpdir to a str for the test case
    save_data_folder = str(tmpdir)

    # Test case 1: Write a single JSON
    json_data1 = {'param1': [1, 2, 3], 'param2': ['a', 'b', 'c']}
    file_name1 = 'file1.json'
    to_save = [(json_data1, file_name1)]
    write_json_locally(to_save, save_data_folder)

    # Check if the file was written
    file_path1 = os.path.join(save_data_folder, file_name1)
    assert os.path.exists(file_path1)
    assert os.path.isfile(file_path1)

    # Read the content of the written JSON file and compare with the original data
    with open(file_path1, 'r') as f:
        loaded_json1 = json.load(f)

    assert loaded_json1 == json_data1

    # Test case 2: Write multiple JSONs
    json_data2 = {'param3': [4, 5, 6], 'param4': ['x', 'y', 'z']}
    file_name2 = 'file2.json'
    json_data3 = {'param5': [7, 8, 9], 'param6': ['m', 'n', 'o']}
    file_name3 = 'file3.json'
    to_save = [(json_data2, file_name2), (json_data3, file_name3)]
    write_json_locally(to_save, save_data_folder)

    # Check if the files were written
    file_path2 = os.path.join(save_data_folder, file_name2)
    assert os.path.exists(file_path2)
    assert os.path.isfile(file_path2)

    file_path3 = os.path.join(save_data_folder, file_name3)
    assert os.path.exists(file_path3)
    assert os.path.isfile(file_path3)

## test remove_folder_locally ## 
def test_remove_folder_locally(tmpdir):
    # Convert tmpdir to a str for the test case
    folder_to_remove = str(tmpdir)

    # Create some test data in the temporary folder
    file_path1 = os.path.join(folder_to_remove, 'file1.txt')
    file_path2 = os.path.join(folder_to_remove, 'file2.txt')

    with open(file_path1, 'w') as f1, open(file_path2, 'w') as f2:
        f1.write("Test file 1")
        f2.write("Test file 2")

    # Ensure the folder exists before removing
    assert os.path.exists(folder_to_remove)

    # Test the remove_folder_locally function
    remove_folder_locally(folder_to_remove)

    # Check if the folder was removed
    assert not os.path.exists(folder_to_remove)

    # Check if the test data files are not present anymore
    assert not os.path.exists(file_path1)
    assert not os.path.exists(file_path2)

## test prepare_dataframe_to_hf ##
def test_prepare_dataframe_to_hf():
    # Sample input DataFrames
    columns = ['Timestamp', 'Pred_Mean', 'Id_Pred']
    data_pred = {
        'Timestamp': ['2023-07-01', '2023-07-02', '2023-07-03'],
        'Pred_Mean': [10, 20, 30],
        'Id_Pred': ['A', 'B', 'C']
    }

    data_true = {
        'Timestamp': ['2023-07-01', '2023-07-02', '2023-07-03'],
        'y': [8, 18, 25]
    }

    X_pred = pd.DataFrame(data_pred, columns=columns)
    X_true = pd.DataFrame(data_true)

    # Test the prepare_dataframe_to_hf function
    model_name = 'my_model'
    result_df = prepare_dataframe_to_hf(X_true, X_pred, model_name)

    # Check if the resulting DataFrame has the correct columns
    expected_columns = ['ds', 'my_model'.title(), 'y']
    assert list(result_df.columns) == expected_columns

    # Check if the 'my_model' column is correctly populated
    assert list(result_df['my_model'.title()]) == [10, 20, 30]

    # Check if the 'y' column is correctly populated
    assert list(result_df['y']) == [8, 18, 25]

    # Check if 'unique_id' index is set
    assert 'unique_id' not in result_df.columns
    assert result_df.index.name == 'unique_id'

## test mlflow_log_artifact ## 
# Sample data for DataFrames and JSON
df_data = {'A': [1, 2, 3], 'B': [4, 5, 6]}
df = pd.DataFrame(df_data)
json_data = {'param1': [1, 2, 3], 'param2': ['a', 'b', 'c']}

def test_mlflow_log_artifact(tmpdir):
    # Convert tmpdir to a str for the test case
    predict_real_folder = str(tmpdir)

    # Test case 1: Log DataFrames
    experiment_name = 'test_experiment'
    df_filename = 'test_dataframe'
    df_to_save = [(df, df_filename)]

    mlflow_log_artifact(predict_real_folder, experiment_name, df_to_save=df_to_save)

    # Test case 2: Log JSONs
    json_filename = 'test_json'
    json_to_save = [(json_data, json_filename)]

    mlflow_log_artifact(predict_real_folder, experiment_name, json_to_save=json_to_save)

    # Test case 3: Log both DataFrames and JSONs
    df_to_save = [(df, df_filename)] 
    json_to_save = [(json_data, json_filename)]

    mlflow_log_artifact(predict_real_folder, experiment_name, df_to_save=df_to_save, json_to_save=json_to_save)


## test model_tuning ##  
# mock_classes.py
class MockModel:
    def __init__(self, **params):
        self.params = params

    def fit(self, train_df):
        # Mock the model training process
        pass

    def predict(self, validation_df):
        # Mock the model prediction process
        return pd.DataFrame({'predictions': [1, 2, 3], 'Id_Pred': ['a', 'b', 'c']})

class MockEvaluator:
    @staticmethod
    def make_evaluation(df_real, df_pred, id_pred, verbose):
        # Mock the evaluation process and return a sample evaluation DataFrame
        return pd.DataFrame({'RMSE': [0.1], 'MAE': [0.2], 'Percentage Coverage': [0.3], 'Id_Pred': ['AAA']})

def test_model_tuning():
    # Sample data
    train_data = {'A': [1, 2, 3], 'B': [4, 5, 6]}
    validation_data = {'A': [7, 8, 9], 'B': [10, 11, 12]}
    train_df = pd.DataFrame(train_data)
    validation_df = pd.DataFrame(validation_data)

    # Grid parameters for testing
    grid_params = {'param1': [3], 'param2': [False], 'param3': [0.2]}
    best_params = model_tuning(MockModel, MockEvaluator, 'Id_Pred', grid_params, train_df, validation_df)
    reference_best_params = {'param1': 3, 'param2': False, 'param3': 0.2}
    
    assert best_params == reference_best_params

   




   


import os
import mlflow
import pandas as pd
import json
import numpy as np

from ..utility.tele_logger import logger
from ..utility.resources import log_exception, get_module_and_function
from ..utility.check_utils import check_not_in_iterable, check_not_isinstance
from ..algorithms.models import HierarchicalReconcileModel


@log_exception(logger)
class JobClass():
    def __init__(self, model_name: str, time_granularity: str, mlflow_runs: pd.DataFrame, apply_hierarchical: bool):
        """
        Init of JobClass

        Parameters
        ----------
        model_name : str
            Represents the name of the model which predictions must be reconciled
        time_granularity : str
            Time aggregation on which the date column is converted
        mlflow_runs : pd.DataFrame
            The selected runs loaded from mlflow
        apply_hierarchical : bool
            Boolean value which specify if the hierarchical reconciliation is applied
        """
        
        self.model_name = model_name
        self.runs = mlflow_runs
        self.apply_hierarchical = apply_hierarchical
        self.time_granularity = time_granularity


    def find_best_run(self, decision_function: str):
        """
        Returns the best model from a set of experiments given two metrics

        Parameters
        ----------
        decision_function : str
            Decision function to minimize in order to find the best parameters.

            The Decision function can use ONLY the evaluation parameters, beware that if an evaluation parameter
            has a space " " in its name, inside the evaluation it MUST be written as a "_", for example if using half
            Percentage Coverage and half RMSE the function will be 'percentage_coverage*0.5 + rmse*0.5'
        """
        logger.info("Parsing and Selecting MLFlow runs.")
        slicer =[]
        for idx, row in self.runs.iterrows():
            try:
                num_models = len(json.loads(row['tags.mlflow.log-model.history']))
                if self.apply_hierarchical and num_models == 2:
                    slicer.append(idx)
                elif not self.apply_hierarchical and num_models == 1:
                    slicer.append(idx)
            except:
                logger.warning(f'An Error has been encountered while reading run {row.run_id} for experiment {row.experiment_id}. Skipping the run.')
        
        if len(slicer) != 0:
            sliced_runs = self.runs.iloc[slicer]
        else:
            raise Exception('Trying to get a best run, but none available are found in the runs history. Check the if the model you are trying to use to predict has ever been executed before.')  

        logger.info("Evaluating Best MLFlow run.")
        min = np.inf
        best_run = None
        metrics = [c for c in sliced_runs.columns if c.startswith('metrics.')]
        for idx, row in sliced_runs.iterrows():
            for metric in metrics:
                if self.apply_hierarchical:
                    exec(f'{metric.split(".")[1].replace(" ","_").lower().split(" ")[0]}=row[metric]')
                else:
                    exec(f'{metric.split(".")[1].replace(" ","_").lower()}=row[metric]')

            score = eval(decision_function)
            if score < min:
                min = score
                best_run = row

        # Load model
        logger.info("Retrieving best Model.")

        check_not_isinstance(obj=best_run, data_type=pd.Series, func=get_module_and_function(), attached_message='Something went wrong. The retrieved best run inside mlflow is not a pandas Series')
        model_uri = f"runs:/{best_run['run_id']}/{self.model_name}.pkl"
        self.Model = mlflow.sklearn.load_model(model_uri)

        if self.apply_hierarchical:
            folder_best_reconcile = 'best_reconciler'
            best_reconciler = os.listdir(os.path.join(best_run.artifact_uri, folder_best_reconcile))[0]
            reconciler_uri = f"runs:/{best_run['run_id']}/{folder_best_reconcile}/{best_reconciler}"
            self.best_reconciler = mlflow.sklearn.load_model(reconciler_uri)

            path_artifacts = os.path.join(best_run.artifact_uri, 'dataframes')
            with open(os.path.join(path_artifacts, 'tags.json'), 'r') as f:
                tmp = json.load(f)

            valid_true = pd.read_parquet(os.path.join(path_artifacts, 'valid_real.parquet'))
            valid_pred = pd.read_parquet(os.path.join(path_artifacts, 'valid_pred.parquet'))
            test_true  = pd.read_parquet(os.path.join(path_artifacts, 'test_real.parquet'))
            test_pred  = pd.read_parquet(os.path.join(path_artifacts, 'test_pred.parquet'))

            self.X_true = pd.concat([valid_true, test_true])
            self.X_pred = pd.concat([valid_pred, test_pred])

            self.S = pd.read_parquet(os.path.join(path_artifacts, 'summing_matrix.parquet'))
            self.tags = {k: np.array(v,dtype=object) for k,v in tmp.items()}


    def make_future_dataset(self, start_date: str, num_forecast_periods: str):
        """
        Generates a dataframe on which the predictions are applied starting from a given date

        Parameters
        ----------
        start_date : str
            The starting date
        num_forecast_periods : str
            The number of forecast periods to generate

        Returns
        -------
        pd.DataFrame
            The dataframe with future values
        """

        logger.info("Generating Future data.")

        dates = pd.date_range(start = start_date, periods = num_forecast_periods, freq = self.time_granularity)
        id_preds = list(self.Model.models_dict.keys())

        future_df = pd.DataFrame(columns=['ds', 'unique_id'])
        for id_pred in id_preds:
            tmp = pd.DataFrame({"ds": dates, 'unique_id':[id_pred]*num_forecast_periods})
            future_df = pd.concat([future_df, tmp])

        self.future_df = future_df.set_index('unique_id')
    
    
    def predict(self, start_date: str, num_forecast_periods: str, decision_function: str ='rmse*0.5 + mae*0.5') -> pd.DataFrame:
        """
        Method for generate future predictions

        Parameters
        ----------
        start_date : str
            The starting date
        num_forecast_periods : str
            The number of forecast periods to generate
        decision_function : str
            Decision function to minimize in order to find the best parameters, by default 'rmse*0.5 + mape*0.5'

            The Decision function can use ONLY the evaluation parameters, beware that if an evaluation parameter
            has a space " " in its name, inside the evaluation it MUST be written as a "_", for example if using
            Percentage Coverage instead of the rmse inside the default function, this becomes 'percentage_coverage*0.5 + mape*0.5'
        
        Returns
        -------
        pd.DataFrame
            Output dataframe
        """

        logger.info("START Predicting Future data.", important=True)

        self.find_best_run(decision_function=decision_function)
        self.make_future_dataset(start_date=start_date, num_forecast_periods=num_forecast_periods)
        forecast_model = self.Model.predict(self.future_df)

        if self.apply_hierarchical:
            try:
                reconcile = f'{self.best_reconciler.__class__.__name__}(method="{self.best_reconciler.method}")'
            except AttributeError:
                reconcile = f'{self.best_reconciler.__class__.__name__}()'
            
            logger.info(f'Applying Hierarchical Reconciliation Model: {reconcile}', important=True)

            HRM = HierarchicalReconcileModel(model_name=self.model_name, S=self.S, tags=self.tags, reconciler=reconcile)
            forecast_rec = HRM.fit_predict(X_true=self.X_true, X_pred=self.X_pred, X_future_pred=forecast_model, as_output=True)

            return forecast_model, forecast_rec
        else:
            return forecast_model, None


from __future__ import annotations

import itertools
import os
import shutil
from typing import Any, Callable, Dict, List, Tuple, Type, Union
import ast
import pandas as pd
import numpy as np
import mlflow
import json
import concurrent.futures

from ..datahandler.datahandler import DataHandler
from ..utility.check_utils import check_not_isinstance, check_not_in_iterable
from ..utility.resources import get_module_and_function, log_exception
from ..utility.tele_logger import logger
from ..analytics.evaluationMetrics import EvaluateModel

def train_val_test_split(X: pd.DataFrame, val_size : int, test_size : int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Executes the splits on the given preprocessed Dataframe.

    Parameters
    ----------
    X : pd.DataFrame
        Preprocessed pd.DataFrame, it needs to be preprocessed by datapreparation.pred.PreprocessingClass beforehand
    val_size : int
        Number of observations based on frequecy for Validation set
    test_size : int
        Number of observations based on frequecy for Test set

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]
        Returns three pd.DataFrames respectively, Train set, Validation set and Test set
    """
    # Parameters check
    #check_split_input(df = X, val_size = val_size, test_size = test_size)

    train_df = X.reset_index()

    test_df = train_df.groupby('unique_id').tail(test_size)
    train_df = train_df.drop(test_df.index)

    valid_df = train_df.groupby('unique_id').tail(val_size)
    train_df = train_df.drop(valid_df.index)

    test_df = test_df.set_index('unique_id')
    valid_df = valid_df.set_index('unique_id')
    train_df = train_df.set_index('unique_id')

    return train_df, valid_df, test_df

def literal_evaluation(values_list: List[str]) -> List[Any]:
    """
    The function evaluates the list values for the parameters: 'weekly_seasonality', 'yearly_seasonality'

    Parameters
    ----------
    values_list : List[str]
        List of strings to be evaluates

    Returns
    -------
    List[Any]
        List of evaluated values
    """
    evaluated_params = []
    for val in values_list:
        try:
            val = ast.literal_eval(val)
        except:
            pass
        evaluated_params.append(val)
    return evaluated_params

def generate_values(dictionary: Dict[str, Dict[str, Union[int, float]]], default_step: Union[int, float]) -> List[Union[int, float]]:
    """
    The function generates the possible values of the numeric prophet parameters

    Parameters
    ----------
    dictionary : Dict[str, Dict[str, Union[int, float]]]
        Dictionary format he config file with min, max and step
    default_step, Union[int, float]
        The default step used if it's not specified in the config file

    Returns
    -------
    List[Union[int, float]]
        List of possible values

    Raises
    ------
    Exception
        If the parameter is missing a key among "min", "max" and "step"
    ValueError
        If the step is greater than the values range
    """
    
    if len(list(dictionary.keys())) == 3:
        for key in dictionary.keys():
            check_not_in_iterable(obj=key, iterable=['min', 'max', 'step'], func=get_module_and_function(),
                                attached_message="The parameter value must be of the following form: {min = <int or float>, max =<int or float>, step = [<int or float>, -1, 'None']}")
    else:
        raise Exception("The parameter value must be of the following form: {min = <int or float>, max =<int or float>, step = [<int or float>, -1, 'None']}")

    min_value = dictionary.get('min')
    max_value = dictionary.get('max')
    step = dictionary.get('step')
    if max_value <= min_value:
        raise ValueError(f'The max_value {max_value} must be greater than the min_value {min_value}')

    if step == 'None':
        step = default_step  # Use Default value as step

    check_not_isinstance(obj=min_value, data_type=(float, int), func=get_module_and_function())
    check_not_isinstance(obj=max_value, data_type=(float, int), func=get_module_and_function())
    check_not_isinstance(obj=step, data_type=(float, int), func=get_module_and_function())

    if step == -1 and min_value > 0:
        step = min_value  # Use min value as step
    elif step == -1 and min_value == 0:
        step = default_step
    elif step > (max_value - min_value):
        raise ValueError(f'The step {step} must be lower than the difference between max {max_value} and min {min_value}')

    values = list(np.arange(min_value, max_value, step))

    # Checks if max_value is already inside the grid
    if max_value not in values:
        values.append(max_value)

    return values

@log_exception(logger)
def grid_values_hyperparameters(config_hyperparameters: Dict[str, Union[List[str], Dict[str, Union[int, float]]]],
                                default_steps: Dict[str, Union[int, float]]) -> Dict[str, List[Union[str, bool, int, float]]]:
    """
    The function takes as input the dictionary of the config file
    and generates the combination of grid search parameters

    Parameters
    ----------
    config_hyperparameters: Dict[str, Union[Dict[str, Union[int, float, str]], List[str]]]
        The dictionary of hyperparameters from the config file
    default_steps: Dict[str, Union[int, float]]
        The default steps used if it's not specified in the config file

    Returns
    -------
    Dict[str, List[Union[str, bool, int, float]]]
        The grid search parameter dictionary
    """
 
    grid_parameters = {}
    for param in config_hyperparameters.keys():
        if type(config_hyperparameters[param]) == list:
            grid_parameters[param] = literal_evaluation(config_hyperparameters[param])

        elif type(config_hyperparameters[param]) == dict:
            check_not_in_iterable(obj=param, iterable=list(default_steps.keys()), func=get_module_and_function(),
                                  attached_message=f'Parameter {param} do NOT have a default setting, add it inside the config/default_steps.toml')

            grid_parameters[param] = generate_values(config_hyperparameters[param], default_step = default_steps[param])

        else:
            raise ValueError(f' The value datatype of {param} is not valid')
    return grid_parameters

def model_tuning(model: Type[Callable], evaluator: Type[EvaluateModel], unique_id: str,
                 grid_params: Dict[str, List[Union[str, bool, int, float]]], 
                 train_df: pd.DataFrame, validation_df: pd.DataFrame,
                 decision_function: str = 'rmse*0.5 + mae*0.5') -> Dict[str, Union[str, bool, int, float]]:
    """
    Performs the parameters selection on a given set of parameters grid

    Parameters
    ----------
    model : TypeType[Callable]
        Model to use in order to find its best parameters
    evaluator : Type[EvaluateModel]
        EvaluateModel class to use for the evaluations
    grid_params : Dict[str, List[Union[str, bool, int, float]]]
        The parameters grid space
    unique_id : str
        Hierarchys ID to be evaluated 
    train_df : pd.DataFrame
        The dataframe on which the model fit is applied 
    validation_df : pd.DataFrame
        The dataframe on which the model evaluation is applied
    decision_function : str, optional
        Decision function to minimize in order to find the best parameters, by default 'rmse*0.5 + mape*0.5'

        The Decision function can use ONLY the evaluation parameters, beware that if an evaluation parameter
        has a space " " in its name, inside the evaluation it MUST be written as a "_", for example if using
        Percentage Coverage instead of the rmse inside the default function, this becomes 'percentage_coverage*0.5 + mape*0.5'

    Returns
    -------
    Dict[str, Union[str, bool, int, float]]
        Dictionary of the best parameters

    Raises
    ------
    ValueError
        If the validation set is empty, after cleaning
    """  

    ######## FUNCTION TO PARALLELIZE ######## 
    def evaluate_params(params: Dict[str, Union[str, bool, int, float]], decision_func: str) -> float:
        """
        Evaluates the given parameter set.

        Parameters
        ----------
        params : Dict[str, Union[str, bool, int, float]]
            Dictionary of parameter values
        decision_function : str
            Decision function to minimize in order to find the best parameters

        Returns
        -------
        float
            Result of the evaluation
        """        

        m =  model(**params)
        m.fit(train_df)

        validation_df_clean = validation_df.dropna()
        if len(validation_df_clean) > 0:
            pred = m.predict(validation_df_clean)
            evals = evaluator.make_evaluation(df_real=validation_df_clean, df_pred=pred, id_pred=unique_id, verbose=False)

            for k in evals.drop(columns='Id_Pred').to_dict().keys():
                exec(f"{k.lower().replace(' ', '_')} = {evals[k][0]}")

            return {eval(decision_function.lower()): params}
        else:
            raise ValueError("After NaNs removal, Validation set is empty.")


    get_minimum = {}
    keys, values = zip(*grid_params.items())

    permutations_dicts = []
    for v in itertools.product(*values):
        if not isinstance(v, list):
            v = list(v)
        
        permutations_dicts.append(dict(zip(keys, v)))

    # Use ThreadPoolExecutor for parallel execution
    with concurrent.futures.ThreadPoolExecutor() as executor:
        # Map evaluate_params function to each parameter set in parallel
        results = executor.map(lambda params: evaluate_params(params, decision_function), permutations_dicts)
        # Collect the results

        for res in results:
            get_minimum.update(res)

    # Find the best parameters
    minimum_error = np.min(list(get_minimum.keys()))
    best_params = get_minimum[minimum_error]

    return best_params

def create_zero_dataframe(columns: List[str], n_rows: int) -> pd.DataFrame:
    """Creates a dataframe with the given columns and rows full of zeros.

    Parameters
    ----------
    columns : List[str]
        Column names
    n_rows : int
        Number of rows

    Returns
    -------
    pd.DataFrame
        dataframe
    """
    # Create a dictionary with the column names and values
    data = {col: np.zeros(n_rows).astype(int) for col in columns}

    # Create the dataframe
    df = pd.DataFrame(data)

    return df


def mlflow_log_artifact(predict_real_folder: str, experiment_name: str, 
                        df_to_save: List[Tuple[pd.DataFrame, str]] =None,
                        json_to_save: List[Tuple[Dict[str, List[str]], str]] =None):
    """The function logs, if provided, the dataframes and the jsons on MLFlow

    Parameters
    ----------
    predict_real_folder : str
        The local folder to temporary save the data
    experiment_name : str
        The name of the experiment
    df_to_save : List[Tuple[pd.DataFrame, str]], optional
        The list of tuples (dataframes, filename) to save , by default None
    json_to_save : List[Tuple[Dict[str, List[str]], str]], optional
        The list of tuples (dict, filename) to save, by default None
    """
        
    logger.info('START logging results.', important=True)

    save_data_folder = os.path.join(predict_real_folder, experiment_name)

    # If there are dataframes to save, apply the function
    if isinstance(df_to_save, list):
        format_ = '.parquet'
        df_to_save = [(D, N+format_) for D, N in df_to_save]

        write_dataframes_locally(to_save=df_to_save, save_data_folder=save_data_folder)

        for _, filename in df_to_save:
            mlflow.log_artifact(os.path.join(save_data_folder, filename), artifact_path="dataframes")

    # If there are dictionaries to save, apply the function
    if isinstance(json_to_save, list):
        format_ = '.json'
        json_to_save = [(D, N+format_) for D, N in json_to_save]

        write_json_locally(to_save=json_to_save, save_data_folder=save_data_folder)
        
        for _, filename in json_to_save:
            mlflow.log_artifact(os.path.join(save_data_folder, filename), artifact_path='dataframes')

    # If dataframes or dictionaries have been saved, apply the function
    if isinstance(df_to_save, list) or isinstance(json_to_save, list):
        remove_folder_locally(save_data_folder)

    logger.info('DONE logging results.', important=True)


def write_dataframes_locally(to_save: List[Tuple[pd.DataFrame, str]], save_data_folder: str):
    """Writes locally the dataframes needed by MLFlow in order to log them in its application.

    Parameters
    ----------
    to_save : List[Tuple[pd.DataFrame, str]]
        List of Tuples with pandas DataFrames and corresponding filenames that will be written
    save_data_folder : str
        Data folder where files will be written
    """
    os.makedirs(save_data_folder, exist_ok=True)
    DH = DataHandler(data_folder = save_data_folder)
    for D, N in to_save:
        DH.write(D, filename = N, folder = '')

def write_json_locally(to_save: List[Tuple[Dict[str, List[str]], str]], save_data_folder: str):
    """Writes locally the jsons needed by MLFlow in order to log them in its application.

    Parameters
    ----------
    to_save : List[Tuple[Dict[str, List[str]], str]]
        List of Tuples with Dictionaries and corresponding filenames that will be written
    save_data_folder : str
        Data folder where files will be written
    """
    os.makedirs(save_data_folder, exist_ok=True)
    for J, N in to_save:
        json_file = os.path.join(save_data_folder, N)
        tmp = {k: list(v) for k,v in J.items()}
        with open(json_file, 'w') as f:
            json.dump(tmp, f)


def remove_folder_locally(folder_to_remove: str):
    """Deletes the folder once the files are logged on MLFlow.

    Parameters
    ----------
    folder_to_remove : str
        Folder to be removed
    """
    shutil.rmtree(folder_to_remove)

def prepare_dataframe_to_hf(X_true: pd.DataFrame, X_pred: pd.DataFrame, model_name: str) -> pd.DataFrame:
    """Prepare the predicted Dataframe from a model from algorithms.models to be reconciled by one
    of HierarchicalForecast reconcilers.

    Parameters
    ----------
    X_true : pd,DataFrame
        Pandas Dataframe containing the true data
    X_pred : pd.DataFrame
        Pandas Dataframe containing the predictions of a model from algorithms.models
    model_name : str
        Name of the model applied to the X_pred

    Returns
    -------
    pd.DataFrame
        Pandas Dataframe ready to be give to a HierarchicalForecast reconciler
    """
    check_not_in_iterable(obj='Timestamp', iterable=X_pred.columns, func=get_module_and_function())
    check_not_in_iterable(obj='Pred_Mean', iterable=X_pred.columns, func=get_module_and_function())
    check_not_in_iterable(obj='Id_Pred', iterable=X_pred.columns, func=get_module_and_function())

    hier_pred = X_pred.rename(columns = {
        'Timestamp' : 'ds', 'Pred_Mean' : model_name.title(), 'Id_Pred': 'unique_id'
    })

    hier_pred = hier_pred.set_index('unique_id')
    for col in hier_pred.columns:
        if col not in ['ds', model_name.title()]:
            hier_pred = hier_pred.drop(col, axis = 1)
    
    hier_pred['y'] = list(X_true['y'])

    return hier_pred


def to_category(
    df: pd.DataFrame, columns: list[str], inplace: bool = False
) -> pd.DataFrame:
    """
    Convert the specified columns in a pandas DataFrame to categorical data type.

    Parameters
    ----------
    df : pandas.DataFrame
        The input DataFrame to be converted.
    columns : List[str]
        The list of column names to be converted to categorical data type.
    inplace : bool, optional
        If True, modify the input DataFrame in place. Otherwise, return a new copy of the DataFrame.
        The default is False.

    Returns
    -------
    pandas.DataFrame
        The modified DataFrame with the specified columns converted to categorical data type.
        If inplace=True, the original DataFrame is modified.
    """

    if not inplace:
        df = df.copy()

    for c in columns:
        df[c] = df[c].astype("category")

    return df


import holidays
import pandas as pd
import numpy as np
from statsmodels.tsa.seasonal import seasonal_decompose
from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self
from typing import List, Tuple, Union, Dict
from hierarchicalforecast.utils import aggregate


from ..utility.tele_logger import logger
from ..utility.check_utils import check_not_isinstance, check_not_in_iterable, check_datatype_convertible,check_dates_complete
from ..utility.resources import get_module_and_function
from ..utility.constants import code_to_region_name, code_to_speciality

################# Transformer Functions #################
def map_columns(hierarchy: Dict[str, str], df: pd.DataFrame, 
                conversion: Dict[str, str]) -> pd.DataFrame:
    
    """
    Change the columns in the dataframe <df> according to a given <hierarchy> and <conversion> dictionary.
    The hierarchy and the conversion dictionaries MUST be specifed in the configuration file.

    Parameters
    ----------
    hierarchy : Dict[str, str]
        A dictionary mapping levels to the <df> column names.
    df : pd.DataFrame
        The input DataFrame to be mapped.
    conversion : Dict[str, str]
        A dictionary mapping levels to conversion names.

    Returns
    -------
    pd.DataFrame
        Containing the modified dataframe if there is a conversion to apply
    """
    # Verify that hierarchy is of type dict
    check_not_isinstance(obj = hierarchy, data_type = dict, func = get_module_and_function())
    # Verify that conversion is of type dict
    check_not_isinstance(obj = conversion, data_type = dict, func = get_module_and_function())
    # Verify that df is of type pd.DataFrame
    check_not_isinstance(obj = df, data_type = pd.DataFrame, func = get_module_and_function())

    df = df.copy()

    for conversion_level in conversion.keys():
        if conversion[conversion_level]:
            name_column = hierarchy[conversion_level]

            # Verify that name_column is a column in df
            check_not_in_iterable(obj = name_column, iterable = df.columns, func = get_module_and_function())

            dict_mapping = eval(conversion[conversion_level]) # Takes the conversion level string and reads it from utility.constants
            df[name_column] = df[name_column].apply(lambda x: dict_mapping[x])
    return df


def column_to_int(hierarchy: Dict[str, str], df: pd.DataFrame, 
                conversion: Dict[str, str]) -> pd.DataFrame:
    
    """
    Change datatypes in the columns in the dataframe <df> according to a given <hierarchy> and <conversion> dictionary.
    The hierarchy and the conversion dictionaries MUST be specifed in the configuration file.

    Parameters
    ----------
    hierarchy : Dict[str, str]
        A dictionary mapping levels to the <df> column names.
    df : pd.DataFrame
        The input DataFrame to be mapped.
    conversion : Dict[str, str]
        A dictionary mapping levels to conversion names. 
        If a conversion name is provided for a level, the corresponding column in the DataFrame will be converted to int64.

    Returns
    -------
    pd.DataFrame
        Containing the modified dataframe if there is a conversion to apply
    """
    # Verify that hierarchy is of type dict
    check_not_isinstance(obj = hierarchy, data_type = dict, func = get_module_and_function())
    # Verify that conversion is of type dict
    check_not_isinstance(obj = conversion, data_type = dict, func = get_module_and_function())
    # Verify that df is of type pd.DataFrame
    check_not_isinstance(obj = df, data_type = pd.DataFrame, func = get_module_and_function())

    df = df.copy()

    for conversion_level in conversion.keys():
        if conversion[conversion_level]:
            name_column = hierarchy[conversion_level]
            for column_val in df[name_column]:
                check_datatype_convertible(obj = column_val, data_type_conversion = int, func = get_module_and_function())
            df[name_column] = df[name_column].astype(np.int64)
    return df

def generate_time_series(df: pd.DataFrame, date_col: str,
                         time_granularity: str, hierarchy: Dict[str, str]) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]:
    """
    Sequential execution of transformations to obtain a DataFrame with a time series structure

    Parameters
    ----------
    df : pd.DataFrame
        Raw dataframe from which generates timeseries
    date_col : str
        Column name identifying the columns to index on
    time_granularity : str 
        Specifies temporal granularity
    hierarchy : Dict[str, str]
        A dictionary mapping levels to the <df> column names.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]
        Returns a tuple with the following elements:
            - Y_df dataframe with the hierarchies as indexes
            - S_df binary matrix to understand the hierarchy levels
            - tags maps the hierarchy level to the hierarchy values 
    """
    
    # Verify that df is of type pd.DataFrame
    check_not_isinstance(obj = df, data_type = pd.DataFrame, func = get_module_and_function())
    # Verify that date_col is a column in df
    check_not_in_iterable(obj = date_col, iterable = df.columns, func = get_module_and_function())
    # Verify time granularity date type
    check_not_isinstance(obj = time_granularity, data_type = str, func = get_module_and_function())
    # Verify hierarchy date type
    check_not_isinstance(obj = hierarchy, data_type = dict, func = get_module_and_function())

    # Raw dataset from which to generate the time series
    df = df.copy()
    
    # Starting generating the time series
    hier_df = get_hierarchical_df(df = df, hierarchy = hierarchy, time_granularity = time_granularity, date_col = date_col)
    Y_df, S_df, tags = get_hierarchical_info(hier_df = hier_df)

    Y_df['y'] = Y_df['y'].replace(0, np.nan)

    return Y_df, S_df, tags

def get_hierarchical_df(df: pd.DataFrame, hierarchy: Dict[str, str],
                        time_granularity: str, date_col: str) -> pd.DataFrame:
    """
    The function generates the dataframe suitable for obtaining hierarchical info. 
    It generates the dataframe starting from the input dataframe by using the information
    of the configuration file.

    Parameters
    ----------
    df : pd.DataFrame
        The input dataframe (typically inside .../data/input)
    hierarchy : Dict[str, str]
        Dictionary retrieved from the configuration file.
        it contains the hierarchy columns ordered by levels
    time_granularity : str
        String referring to the time granularity.
        Information retrieved from the configuration file.
    date_col : str
        The string referring to the date column of the dataframe
    Returns
    -------
    pd.DataFrame
        A dataframe of columns: ds|y|level0|level1|level2|level3
        level0 refers to Italia
        level1, level2, level3 are obtained from the hierarchy dict
    """
    
    output_df = df[[date_col]] # Grabs the time columns, output_df remains a dataframe
    output_df['level0'] = 'Italia' # Sets level0 as the macro hierarchy Italia
    output_df['y'] = 1 # Sets 1 for each observation

    # Transforms the dataframe based on a given time_granularity (e.g. 'H' (hours), 'D' (days), 'M' (months), ...)
    # and clips them to the said granularity in order to have multiple same dates to let the aggregate sum them up
    output_df[date_col] = output_df[date_col].dt.to_period(time_granularity).dt.to_timestamp()

    # For each level, attaches the column to the output_df
    for level, col_name in hierarchy.items():
        output_df[level] = df[col_name]
    output_df = output_df.rename(columns={date_col:'ds'}) # Normalize the date_col name into 'ds'
    return output_df

def get_hierarchical_info(hier_df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]:
    """
    The function generates the hierarchical aggregation info
    starting from the hierarchical dataframe obtained by <get_hierarchical_df>

    Parameters
    ----------
    hier_df : pd.DataFrame
        Dataframe obtained by <get_hierarchical_df>

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]])
        Returns a tuple with the following elements:
            - Y_df dataframe with the hierarchies as indexes
            - S_df binary matrix to understand the hierarchy levels
            - tags maps the hierarchy level to the hierarchy values 

    """
    # Filters and sorts the columns regarding hiherarchy's levels
    filtered_columns = [col for col in hier_df.columns if 'level' in col]
    filtered_columns.sort()

    spec = [filtered_columns[:i] for i in range(1, len(filtered_columns) + 1)] # Builds the spec needed for the aggregate function
    Y_df, S_df, tags = aggregate(df = hier_df, spec = spec) # Executes the aggregations based on timestamps

    return Y_df, S_df, tags

def missing_data_imputation(missing_data_strategy: Union[str,int,dict], df: pd.DataFrame) -> pd.DataFrame:
    """
    The function imputes the missing values of the dataframe <df> with a given strategy <missing_data_strategy>

    Parameters
    ----------
    missing_data_strategy : dict, str or int
        Identifies whether to impute missing values and if so using which strategy/value
        Allowed parameters: 
            if str:
                ""      : none of the missing values are replaced
                "mean"  : missing values are replaced with the mean of the known values in the dataset.
                "median": missing values are replaced with the median of the known values in the dataset.
                "zero"  : missing values are replaced with the 0.
                "bfill" : missing values are replaced with the next available value in the dataset.
                "ffill" : missing values are replaced with the most recent preceding value in the dataset.
            if int:
                replace NaN with the specified integer
            if dict:
                replace NaN using the specified interpolation method (allowed "polynomial" or "spline") and its order.

    df : pd.DataFrame
    The dataframe which target column will be filled with a given strategy

    Returns
    -------
    pd.DataFrame
        The dataframe afer the imputation

    Raises
    ------
    ValueError
        Invalid value for spline order
    """

    df = df.copy()
    if isinstance(missing_data_strategy, dict): 
        if missing_data_strategy['interpolation'] == "spline" and (df['y'].notna().sum() < missing_data_strategy['order'] or missing_data_strategy['order'] >5):
            logger.error('The number of data points must be larger than the spline degree k or k should be 1 <= k <= 5.')
            raise ValueError("Invalid value for spline order in 'datapreparation_utils.generate_time_series'")
        
    if missing_data_strategy == "":
        logger.info('No Missing Data Imputation applied.')
    else:
        logger.info(f'Missing Data Imputation with Strategy: {missing_data_strategy}')
        fillna = ReplacerNA(missing_data_strategy)
        df = fillna.fit_transform(df)
        df['y'] = df['y'].clip(lower=0)

    return df

def completing_calendar(df: pd.DataFrame, time_granularity: str) -> pd.DataFrame:
    """
    Fill missing dates in the given DataFrame 'df' to complete the calendar,
    considering the specified 'time_granularity'.

    Parameters:

    df: pd.DataFrame
    The input DataFrame containing a temporal column 'ds' and a value column 'y'.
    
    time_granularity: str 
    The frequency level to consider for completing the calendar (e.g., 'D' for daily, 'H' for hourly, 'W' for weekly, etc.).

    Returns:
        pd.DataFrame: The DataFrame with the calendar completed, with missing dates filled with NaNs (or 0s).
    """

    df = df.copy()
    index_col = df.index.name
    ds, y = df.columns

    # Check if the dates are already complete at the specified granularity, if not refill df
    if not check_dates_complete(date_series=df[ds], freq=time_granularity):

        df = df.reset_index()
        df.set_index(ds, inplace=True)
        # Pivot the DataFrame to create a complete calendar at the specified granularity
        df = df.pivot(columns=index_col, values=y).resample(time_granularity).sum()
        # Reset the index and melt the DataFrame to bring it back to the original format
        df = df.reset_index().melt(id_vars=ds, var_name=index_col, value_name=y)
        df = df.sort_values(by=[index_col, ds]).set_index(index_col)
        df[y] = df[y].replace(0, np.nan)

    return df

################# Utils Teleconsulto  #################
def filter_target_col(df: pd.DataFrame, target_col: Union[str, List[str]]) -> pd.DataFrame:
    """
    The function filters the rows of the dataframe on the unique target_col 

    Parameters
    ----------
    df : pd.DataFrame
        The input dataframe
    target_col : Union[str, List[str]]
        The target column

    Returns
    -------
    pd.DataFrame
        The dataframe with the unique values for the target col
    """
    df = df.copy()
    df = df.drop_duplicates(subset=target_col)
    return df

################# Transformer Classes #################
class Normalizer(BaseEstimator, TransformerMixin):
    def __init__(self):
        """
        Normalize data to the range [0, 1].
        """
        pass

    def fit(self, X, y=None):
        """
        Fit the Normalizer to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Normalizer object.
        """
        
        self.min = np.min(X)
        self.max = np.max(X)
        return self

    def transform(self, X):
        """
        Normalize the input data to the range [0, 1].

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_normalized : array-like
            Normalized data obtained by scaling the values to the range [0, 1].
        """
        X_normalized = (X - self.min) / (self.max - self.min)
        return X_normalized

    def inverse_transform(self, X_normalized):
        """
        Reconstruct the original data from the normalized data by applying the inverse transformation.

        Parameters:
        -----------
        X_normalized : array-like
            Normalized data.

        Returns:
        --------
        X : array-like
            Reconstructed data obtained by applying the inverse transformation.
        """
        X = X_normalized * (self.max -self.min) + self.min
        return X
  

class ReplacerNA(TransformerMixin, BaseEstimator):

    def __init__(self, method: Union[str, int, Dict[str, Union[str, int]]]) -> Self:

        """Class for handling of NA

        Parameters
        ----------
        method : Union[str, int, Dict[str, Union[str, int]]]]
            If str specify the method to replace NA value (mean,median,zero), if int specify the value to replace NA value
            If dict specify which interpolation method to use between polynomial and spline and its order
        """
        
        self.method = method

    def fit(self, X: pd.DataFrame) -> Self:

        """Compute value useful for replacing NA

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            Fitted replacer
        """
        
        if self.method == "mean":
            self.value = X.iloc[:,1].mean()
            self.method_for_df = None
        elif self.method == "median":
            self.value = X.iloc[:,1].median()
            self.method_for_df = None
        elif self.method == "zero":
            self.value = 0
            self.method_for_df = None
        elif self.method == "bfill":
            self.value = None
            self.method_for_df = "bfill"
        elif self.method == "ffill":
            self.value = None
            self.method_for_df = "ffill"
        elif self.method == "interpolate":
            self.value = None
            self.method_for_df = "interpolate"
        elif isinstance(self.method, dict):
            self.value = self.method["order"]
            self.method_for_df = self.method["interpolation"].lower()

        else:
            self.value = self.method
            self.method_for_df = None

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform replacement of missing values

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            Transformed time series
        """
        if self.method_for_df in ["polynomial", "spline"]:
            # Create a temporary DataFrame with a DatetimeIndex
            temp_df = pd.DataFrame({X.columns[1]: X.iloc[:, 1]})
            temp_df.index = pd.to_datetime(X.iloc[:, 0])

            # Perform time-based interpolation in the temporary DataFrame
            temp_df.iloc[:, 0] = temp_df.iloc[:, 0].interpolate(method=self.method_for_df, order = self.value)

            # Assign the interpolated values to the original column in the X DataFrame
            X.iloc[:, 1] = temp_df.iloc[:, 0].values
        else:
            X.fillna(self.value, method=self.method_for_df, inplace=True)
        return X
    

class Detrender(TransformerMixin, BaseEstimator):

    def __init__(self, period: int) -> Self:

        """Detrending time series

        Parameters
        ----------
        period : int
            Specify period considered for compute additive decomposition

        Returns
        -------
        self : object
        """

        self.period = period


    def fit(self, X: pd.DataFrame) -> Self:

        """Compute additive decomposition useful to detrend time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            Fitted detrender
        """

        additive_decomp = seasonal_decompose(X.iloc[:,1], model="additive", period=self.period, extrapolate_trend="freq")
        self.trend = additive_decomp.trend

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform detrending of time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            Transformed time series
        """

        detrend_time_series = X.iloc[:,1] - self.trend
        ris = pd.concat([X.iloc[:,0],detrend_time_series],axis=1)
        ris.columns = X.columns

        return  ris
    

class Deseasoner(TransformerMixin, BaseEstimator):

    def __init__(self, period: int) -> Self:

        """Deseasonalises time series

        Parameters
        ----------
        period : int
            Specify period considered for compute additive decomposition
        """

        self.period = period


    def fit(self, X: pd.DataFrame) -> Self:

        """Compute additive decomposition useful to deseasonalises time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            Fitted deseasoner
        """
        
        additive_decomp = seasonal_decompose(X.iloc[:,1], model="additive", period=self.period, extrapolate_trend="freq")
        self.seasonal = additive_decomp.seasonal

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform deseasonalises of time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            Transformed time series
        """

        deseason_time_series = X.iloc[:,1] - self.seasonal
        ris = pd.concat([X.iloc[:,0],deseason_time_series],axis=1)
        ris.columns = X.columns

        return ris


class Differencer(TransformerMixin, BaseEstimator):

    def __init__(self, lag: int) -> Self:

        """Differencing time series
        
        Parameters
        ----------
        lag : int
            Differencing time series lag

        Returns
        -------
        self : object
        """

        self.lag = lag

    def fit(self, X: pd.DataFrame) -> Self:

        """Compute value useful to compute differencing time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)         

        Returns
        -------
        self : object
            Fitted normalizer
        """

        self.shape = X.shape[0]
        self.lag_time_series = X.iloc[:self.shape-self.lag,1]
        self.timestamp = X.iloc[self.lag:,0].reset_index(drop=True)

        return self
    
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Perform differencing time series

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing two columns (timestamp and volumes of time series)

        Returns
        -------
        X : pd.DataFrame
            Transformed time series
        """

        time_series_lagged = X.iloc[self.lag:,1].reset_index(drop=True) - self.lag_time_series
        ris = pd.concat([self.timestamp,time_series_lagged], axis=1)
        ris.columns = X.columns
        
        return ris


class OutlierRemover(BaseEstimator, TransformerMixin):
    def __init__(self, lower_threshold_percentile: Union[int, float] =5, upper_threshold_percentile: Union[int, float]=95):
        """
        Remove outliers from a dataset by capping values above and below thresholds.

        Parameters:
        -----------
        lower_threshold_percentile : Union[int, float], optional 
            Percentile threshold below which values will be capped, by default 5
        upper_threshold_percentile : Union[int, float], optional
            Percentile threshold above which values will be capped, by default 95
        """
        self.upper_threshold_percentile = upper_threshold_percentile
        self.lower_threshold_percentile = lower_threshold_percentile
        self.upper_threshold = None
        self.lower_threshold = None

    def fit(self, X, y=None):
        """
        Compute the upper and lower thresholds based on percentiles of the input data.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted OutlierRemover object.
        """
        self.upper_threshold = np.percentile(X, self.upper_threshold_percentile)
        self.lower_threshold = np.percentile(X, self.lower_threshold_percentile)
        return self

    def transform(self, X):
        """
        Cap the values above the upper threshold and below the lower threshold.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_transformed : array-like
            Transformed data with capped outlier values.
        """
        X[X > self.upper_threshold] = self.upper_threshold
        X[X < self.lower_threshold] = self.lower_threshold
        return X
    

class Smoother(BaseEstimator, TransformerMixin):
    def __init__(self, window_size: int):
        """
        Smoothes a time series by applying a moving average window.

        Parameters:
        -----------
        window_size : int
            Size of the moving average window.
        """
        self.window_size = window_size

    def fit(self, X, y=None):
        """
        Fit the Smoother to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Smoother object.
        """
        return self

    def transform(self, X):
        """
        Smooth the input data by applying a moving average window.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_smoothed : array-like
            Smoothed data obtained by applying the moving average window.
        """
        
        X_smoothed = X.rolling(window=self.window_size, min_periods=1).mean()
        return X_smoothed


class Differentiator(BaseEstimator, TransformerMixin):
    def __init__(self, order: int):
        """
        Differentiates "order" times a time series by taking differences between consecutive values.

        Parameters:
        -----------
        order : int
            Order of differentiation.
        """
        self.order = order

    def fit(self, X, y=None):
        
        """
        Fit the Differentiator to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted Differentiator object.
        """
        return self

    def transform(self, X):
        """
        Apply differentiation to the input data by taking differences between consecutive values.
 

        Parameters:
        -----------
        X : array-like
            Input data.


        Returns:
        --------
        X_transformed : array-like
            Transformed data obtained by taking differences between consecutive values.
        """
        X_transformed = X.copy()
        for _ in range(self.order):
            X_transformed = X_transformed.diff()
        X_transformed = np.nan_to_num(X_transformed, nan=0.0)
        return X_transformed


class LogTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        """
        Logarithmic transformation of data.
        """
        pass

    def fit(self, X, y=None):
        """
        Fit the LogTransformer to the data. No computations are needed in this case.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        self : object
            Fitted LogTransformer object.
        """
        return self

    def transform(self, X):
        """
        Apply the logarithmic transformation to the input data.

        Parameters:
        -----------
        X : array-like
            Input data.

        Returns:
        --------
        X_transformed : array-like
            Transformed data obtained by applying the logarithmic transformation.
        """
        X_transformed = np.log1p(X)
        return X_transformed

    def inverse_transform(self, X_transformed):
        """
        Reconstruct the original data from the transformed data by applying the inverse logarithmic transformation.

        Parameters:
        -----------
        X_transformed : array-like
            Transformed data.

        Returns:
        --------
        X_reconstructed : array-like
            Reconstructed data obtained by applying the inverse logarithmic transformation.
        """
        X_reconstructed = np.expm1(X_transformed)
        return X_reconstructed


def datepart(df: pd.Series) -> pd.DataFrame:
    """
    Extracts various date/time parts from a pandas Series of datetime values,
    and returns them as a new DataFrame.

    Parameters
    ----------
    df : pd.Series
        A pandas Series of datetime values to extract date/time parts from.

    Returns
    -------
    pd.DataFrame
        A new pandas DataFrame containing the extracted date/time parts,
        with the same index as the input Series.

    Examples
    --------
    >>> dates = pd.Series(['2022-01-01 12:00:00', '2022-01-02 18:00:00'])
    >>> datepart(dates)
                        dayofmonth  dayofweek  dayofyear  hour  year  month  quarter  holiday  weekend
    2022-01-01 12:00:00           1          5          1    12  2022      1        1     True     True
    2022-01-02 18:00:00           2          6          2    18  2022      1        1    False     True
    """
    it_holidays = holidays.country_holidays('IT')
    df_ds = pd.to_datetime(df.values)
    new_df = pd.DataFrame()
    new_df['dayofmonth'] = df_ds.day
    new_df['dayofweek'] = df_ds.dayofweek
    new_df['dayofyear'] = df_ds.dayofyear
    # new_df['hour'] = df_ds.hour
    new_df['year'] = df_ds.year
    new_df['month'] = df_ds.month
    new_df['quarter'] = df_ds.quarter
    new_df['holiday'] = df_ds.map(lambda x: x in it_holidays)
    new_df['weekend'] = df_ds.dayofweek.map(lambda x: x in [5, 6])
    
    new_df.index = df.values

    return new_df

def get_hierarchy_column(df: pd.DataFrame, data: str, values: str, hier: str, n_hier: int = None, fill_value: int = 0,
                         sons: bool = False, father: bool = False,type_: str=""):
        
    pivot = pd.pivot_table(df.fillna(0),values=values,index=data,columns=hier,fill_value=fill_value)
    pivot.columns = [type_ + "_" + i for i in pivot.columns]
    n_hiers = np.array([n_hier-1,n_hier,n_hier+1])[[father,True,sons]]
    
    if n_hier:
        col = [i for i in pivot.columns if len(i.split("/")) in n_hiers]
        pivot = pivot.loc[:,col]
        
    cols = [list(filter(lambda x : i in x,pivot.columns)) 
            for i in np.unique(np.array([i.rsplit("/",1)[0] for i in pivot.columns if len(i.split("/")) == n_hiers[-1]]))]
    
    return [pivot.loc[:,i] for i in cols]

def divide_by_father(df: pd.DataFrame):
    """
    Divides all columns in a pandas DataFrame by a "father" column,
    and returns the result as a new DataFrame. The father column is
    the one with the least number of "/"

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame containing the columns to be divided,
        as well as a "father" column to divide by.

    Returns
    -------
    pd.DataFrame
        A new pandas DataFrame containing the resulting values of the division.
        The "father" column is included in the new DataFrame.

    Examples
    --------
    >>> data = {'A/B': [1, 2, 3], 'A/C': [4, 5, 6], 'A': [7, 8, 9]}
    >>> df = pd.DataFrame(data)
    >>> divide_by_father(df)
              A/B       A/C    A
    0  0.142857  0.571429  7
    1  0.250000  0.625000  8
    2  0.333333  0.666667  9
    """
    father_index = np.argmin([i.count("/") for i in df.columns])
    father = df.columns[father_index]
    
    df_div = df.div(df[father].values,axis=0)
    df_div[father] = df[father]
    
    df_div.fillna(1/(len(df.columns)-1),inplace=True)
    
    return df_div

def get_lagged_column(df: pd.DataFrame, lags: list[int], columns: list[str] = None):
    """
    Computes lagged versions of one or more columns in a pandas DataFrame,
    and returns them as a new DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame containing the original columns to be lagged.
    lags : list[int]
        A list of integer values indicating the number of lags to compute.
    columns : list[str], optional
        The name of the columns to be lagged. If not specified, all columns in
        the input DataFrame will be lagged.

    Returns
    -------
    pd.DataFrame
        A new pandas DataFrame containing the lagged versions of the specified column(s).

    Examples
    --------
    >>> data = {'A': [1, 2, 3, 4], 'B': [5, 6, 7, 8]}
    >>> df = pd.DataFrame(data)
    >>> get_lagged_column(df, [1, 2], 'A')
       Lag1_A  Lag2_A
    0     NaN     NaN
    1     1.0     NaN
    2     2.0     1.0
    3     3.0     2.0
    """
    if columns:
        cols = [columns]
    else:
        cols = df.columns
        
    col_lagged = [df[cols].shift(lag) for lag in lags]
    col_names = ["_".join(i) for i in itertools.product(["Lag"+str(i) for i in lags],cols)]
    
    df_lag = pd.concat(col_lagged,axis=1)
    df_lag.columns = col_names
    
    return df_lag


import pandas as pd
import numpy as np
from sklearn.metrics import mean_squared_error, r2_score, mean_absolute_error, mean_absolute_percentage_error

from ..utility.check_utils import check_not_isinstance
from ..utility.resources import get_module_and_function
from ..utility.exceptions import ColumnNotFound
from ..utility.tele_logger import logger


class EvaluateModel():
    def __init__(self, date_true: str, date_pred: str, y_true: str, y_pred: str, upper_95: str, lower_95: str):
        """_summary_

        Parameters
        ----------
        date_true : str
            Column name for the timestamps of the real dataset
        date_pred : str
            Column name for the timestamps of the predicted dataset
        y_true : str
            Target column name for the real datasets
        y_pred : str
            Target column name for the predicted datasets
        upper_95 : str
            Column name for Upper CI at 95%
        lower_95 : str
            Column name for Lower CI at 95%
        """
        check_not_isinstance(obj = date_true, data_type = str, func= get_module_and_function())
        check_not_isinstance(obj = date_pred, data_type = str, func= get_module_and_function())
        check_not_isinstance(obj = y_true, data_type = str, func= get_module_and_function())
        check_not_isinstance(obj = y_pred, data_type = str, func= get_module_and_function())
        check_not_isinstance(obj = upper_95, data_type = str, func= get_module_and_function())
        check_not_isinstance(obj = lower_95, data_type = str, func= get_module_and_function())
        self.date_true = date_true
        self.date_pred = date_pred
        self.y_true = y_true
        self.y_pred = y_pred
        self.upper_95 = upper_95
        self.lower_95 = lower_95

    def make_evaluation(self, df_real: pd.DataFrame, df_pred: pd.DataFrame, id_pred: str, verbose: bool =True) -> pd.DataFrame:
        """
        Create a DataFrame representing the evaluation metrics

        Parameters
        ----------
        df_real : pd.DataFrame
            Dataframe containing the real values
        df_pred : pd.DataFrame
            Dataframe containing the predicted values
        id_pred : str
            Hierarchys ID to be evaluated
        suppress_log : bool
            Decide to log or not the messages from the function, by default False

        Returns
        -------
            A pd.DataFrame containing the results for MAE, RMSE, MSE and R2
        """
        if self.date_true not in df_real.columns:
            raise ColumnNotFound(func='evaluations', column=self.date_true, data=df_real)
        if self.date_pred not in df_pred.columns:
            raise ColumnNotFound(func='evaluations', column=self.date_pred, data=df_pred)
        if self.y_true not in df_real.columns:
            raise ColumnNotFound(func='evaluations', column=self.y_true, data=df_real)
        if self.y_pred not in df_pred.columns:
            raise ColumnNotFound(func='evaluations', column=self.y_pred, data=df_pred)

        if verbose:
            logger.info(f'Evaluating results for {id_pred}')

        df_real= df_real.set_index(self.date_true).dropna()
        df_pred= df_pred.set_index(self.date_pred).dropna()
        inter = df_real.index.intersection(df_pred.index)

        if len(inter) == 0:
            logger.warning(f'Length of intersection between Predictions and Ground Truth Datasets for {id_pred} is Null. Probably one of the two Datasets has not available data (all NaN).')
            scores = {'Id_Pred': id_pred, 'MAE':[np.nan], 'MAPE':[np.nan], 'RMSE': [np.nan], 'MSE':[np.nan], 'R2':[np.nan], 'Percentage Coverage':[np.nan]}
        else:    
            df_real = df_real.loc[inter]
            df_pred = df_pred.loc[inter]

            # Mean Absolute Error
            # if verbose:
            #     logger.info('Evaluating MAE')
            mae = round(mean_absolute_error(df_real[self.y_true], df_pred[self.y_pred]),2)
            # Mean Absolute Error
            # if verbose:
            #     logger.info('Evaluating MAPE')
            mape = round(mean_absolute_percentage_error(df_real[self.y_true], df_pred[self.y_pred]),2)
            # Root Mean Square Error
            # if verbose:
            #     logger.info('Evaluating RMSE')
            rmse = round(mean_squared_error(df_real[self.y_true], df_pred[self.y_pred], squared = False),2)
            # Mean Square Error
            # if verbose:
            #     logger.info('Evaluating MSE')
            mse = round(mean_squared_error(df_real[self.y_true], df_pred[self.y_pred]),2)
            # Coefficient of Determination
            # if verbose:
            #     logger.info('Evaluating R2')
            r2 = round(r2_score(df_real[self.y_true], df_pred[self.y_pred]),2)
            # Forecast Coverage Percentage
            # if verbose:
            #     logger.info('Evaluating Forecast Coverage Percentage')
            coverage = self.forecast_coverage(df_real=df_real, df_pred=df_pred)

            scores = {'Id_Pred': id_pred, 'MAE':[mae], 'MAPE':[mape], 'RMSE': [rmse], 'MSE':[mse], 'R2':[r2], 'Percentage Coverage':[coverage]}

            # if verbose:
            #     logger.info(f'Finished Evaluating results for {id_pred}')

        # Saving the scores in a DF
        return pd.DataFrame(scores)
    
    def forecast_coverage(self, df_real : pd.DataFrame, df_pred : pd.DataFrame) -> float:
        """
        Calculate the forecast coverage metric.

        Parameters:
        ----------
        df_real : pd.DataFrame
            DataFrame of real values with columns self.date and self.y_true.
        df_pred : pd.DataFrame
            DataFrame of predicted values with columns self.date, self.y_pred,
            'Pi_lower_95', and 'Pi_upper_95'.

        Returns:
        -------
        Float
            Forecast coverage as a percentage.
        """
        # Check if the actual values fall within the prediction intervals
        within_interval = (df_real[self.y_true] >= df_pred[self.lower_95]) & (df_real[self.y_true] <= df_pred[self.upper_95])
        
        # Calculate the forecast coverage as the percentage of actual values within the prediction intervals
        forecast_coverage = within_interval.mean() * 100
        
        return forecast_coverage



import calendar
import warnings

import matplotlib.pyplot as plt
import numpy as np
from prophet import Prophet
from prophet.plot import plot_components_plotly
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns
from statsmodels.tsa.stattools import acf, adfuller, kpss, pacf

from ..utility.resources import log_exception
from ..utility.tele_logger import logger


class timeSeriesVisualization():

    def __init__(self,df: pd.DataFrame, target: str, data: str, regione: str, use_case: str):

        """Class for time series visualization (plot and test)
        
        Parameters
        ----------
        df : pd.DataFrame
            dataset result of pre-process,containg timeline and target
        target : str
            name of target column
        data : str
            name of timeline column
        regione : str, optional
            name of level hierarchy, by default 
        use_case : str
            use case name
        """

        self.df = df
        self.target = target
        self.data = data
        self.regione = regione
        self.use_case = use_case

    def plotTimeSeries(self) -> tuple:
        
        """Plot and mirror plot of time series

        Parameters
        ----------

        Results
        -------
        tuple
            tuple that containing two figure for plot and mirror plot
        """

        # df = df.dropna()

        # Plot
        fig = px.line(
            self.df,
            x=self.data,
            y=self.target,
            labels={self.data: "data", self.target: f"{self.target}"},
            title=f"Andamento {self.use_case} - {self.regione}",
            color_discrete_sequence=["blue"],
            width=800,
            height=500,
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )

        # Plot (mirror)
        fig_mirror = px.area(
            self.df,
            x=self.data,
            y=self.target,
            labels={self.data: "data", self.target: f"{self.target}"},
            title=f"Andamento {self.use_case} - {self.regione} (mirror plot)",
            color_discrete_sequence=["green"],
            width=800,
            height=500,
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )

        fig_mirror.add_trace(
            go.Scatter(
                x=self.df[self.data],
                y=-self.df[self.target],
                marker=dict(color="green"),
                mode="lines",
                fill="tonexty",
                showlegend=False,
            )
        )

        # fig.show()
        # fig_mirror.show()

        return (fig, fig_mirror)

    def distribution_plot(self) -> plotly.graph_objs._figure.Figure:
        
        """Plot distribution of time series

        Parameters
        ----------

        Returns
        -------
        plotly.graph_objs._figure.Figure
            plot distribution
        """

        fig = px.histogram(
            self.df,
            self.target,
            title=f"Distribuzione di frequenza serie temporale <br> ({self.use_case}-{self.regione})",
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                    bargap=0.2,
                )
            ),
        )

        return fig

    def plot_day_month(self) -> tuple:
        
        """Plot time series aggregated by day and month

        Parameters
        ----------
        
        Returns
        -------
        tuple
            tuple that containing figure for day and month
        """

        time_series = self.df[self.target]
        time_series.index = self.df[self.data]

        # Casi per giorno della settimana
        tmp_abbr = pd.DataFrame.from_dict(
            {a: b for a, b in enumerate(calendar.day_abbr[:])},
            orient="index",
            columns=[self.data],
        )
        tmp_num = time_series.groupby(pd.to_datetime(time_series.index).dayofweek).sum()
        tmp = tmp_abbr.join(tmp_num, how="outer").fillna(0)

        # day = df.groupby(df[data].apply(lambda i: pd.to_datetime(i).dayofweek))[target].sum()
        # day.index = [i for i in calendar.day_abbr]
        # day = day.to_frame().reset_index()
        # day.columns = [data,target]

        fig_day = px.bar(
            tmp,
            x=self.data,
            y=self.target,
            labels={self.data: "Giorno della settimana", self.target: f"Richieste {self.use_case}"},
            title=f"Numero di richieste {self.use_case} per giorno della settimana - {self.regione}",
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )

        # Casi per mese
        tmp_abbr = pd.DataFrame.from_dict(
            {a + 1: b for a, b in enumerate(calendar.month_abbr[1:])},
            orient="index",
            columns=[self.data],
        )
        tmp_num = time_series.groupby(pd.to_datetime(time_series.index).month).sum()
        tmp = tmp_abbr.join(tmp_num, how="outer").fillna(0)

        # month = df.groupby(df[data].apply(lambda i: pd.to_datetime(i).month))[target].sum()
        # month.index = [i for i in calendar.month_abbr[1:]]
        # month = month.to_frame().reset_index()
        # month.columns = [data,target]

        fig_month = px.bar(
            tmp,
            x=self.data,
            y=self.target,
            labels={self.data: "Mese", self.target: f"Richieste {self.use_case}"},
            title=f"Numero di richieste {self.use_case} per mese - {self.regione}",
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )

        # fig_day.show()
        # fig_month.show()

        return (fig_day, fig_month)

    # Test stazionarietà (per individuare stagionalità)
    def stationarity(self) -> str:

        """test to verify stationarity

        Parameters
        ----------

        Returns
        -------
        str
            result of test
        """

        df = self.df.dropna()

        time_series = df[self.target]
        time_series.index = df[self.data]

        dftest = adfuller(time_series, autolag="AIC")
        dfoutput = pd.Series(dftest[0:3], index=["Test Statistic", "p-value", "#Lags Used"])
        DF = dfoutput["p-value"] < 0.05
        res = "Dickey-Fuller Test:\n"
        res += "Se p-value > 0.05: NON-stazionario\n"
        res += "Se p-value < 0.05: stazionario\n"
        res += f"p-value={dfoutput['p-value']}, Stazionario: {DF}\n\n"

        kpsstest = kpss(time_series, regression="ct", nlags="auto")
        kpss_output = pd.Series(
            kpsstest[0:3], index=["Test Statistic", "p-value", "#Lags Used"]
        )
        KPSS = kpss_output["p-value"] >= 0.05
        res += "KPSS Test:\n"
        res += "Se p-value > 0.05: stazionario\n"
        res += "Se p-value < 0.05: NON-stazionario\n"
        res += f"p-value={kpss_output['p-value']}, Stazionario: {KPSS}\n\n"

        if DF == KPSS:
            stazionarieta = DF
        elif DF == False:
            stazionarieta = "trend stazionario"
        else:
            stazionarieta = "differenza stazionaria"
        res += f"Risultato dei test: {stazionarieta}"
        print(res)

        return res

    # Autocorrelation
    def autocorrelation(self, n_lags: int = 50, detrend: bool = True) -> tuple:

        """Plot that show autocorrelation of time series

        Parameters
        ----------
        n_lags : int
            number of lags to compute autocorrelation
        detrend : bool
            whether to detrend the time series before computing the autocorrelation. Default is True

        Returns
        -------
        tuple
            tuple that containing figure for autocorrelation

        """

        if detrend:
            trend = self.additive_decomposition()[0]
            time_series = self.df[self.target] - trend
            title = "Autocorrelazione (detrendizzata)"
        else:
            time_series = self.df[self.target]
            title = "Autocorrelazione"

        autocorr = acf(time_series.dropna(), nlags=n_lags)[1:]
        fig1 = px.bar(
            x=np.arange(autocorr.shape[0]) + 1,
            y=autocorr,
            labels={"x": "Lag", "y": "Autocorrelazione"},
            title=title,
            template=dict(
                layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="red"),
                    font=dict(color="green"),
                )
            ),
        )
        fig1.update_layout(showlegend=False)

        return [fig1]

    # Decomposizione Additiva (Trend + Seasonality + Residual)
    def additive_decomposition(self) -> tuple:

        """Compute and plot additive decompostion

        Parameters
        ----------
        
        Returns
        -------
        tuple
            result of additive decomposition (trend and seasonality)
        """
        df = self.df.dropna()

        df = pd.DataFrame({'ds': df[self.data], 'y': df[self.target]})
        m = Prophet()
        m.fit(df)
        forecast = m.predict(df[["ds"]])

        fig = plot_components_plotly(m, forecast)

        forecast = forecast.set_index("ds")
        try:
            weekly = forecast["weekly"]
        except:
            weekly = None
        return (forecast["trend"], weekly, fig)

    def boxplots_seasons(self) -> plt.figure:
        
        """Creates a boxplot showing weekday-season effects

        Parameters
        ----------
        
        Returns
        -------
        plt.figure
            Pyplot figure in seaborn style
        """

        df_ = self.df.set_index(self.data).copy()
        df_ = df_.copy()
        df_["date"] = df_.index
        df_["weekday"] = df_["date"].dt.day_name()
        df_["month"] = df_["date"].dt.month
        df_["day"] = df_["date"].dt.day
        df_["date_offset"] = (df_.date.dt.month * 100 + df_.date.dt.day - 320) % 1300
        df_["season"] = pd.cut(
            df_["date_offset"],
            [0, 300, 602, 900, 1300],
            labels=["Spring", "Summer", "Fall", "Winter"],
        )
        df_plot = df_[["weekday", "season", self.target]]
        fig, ax = plt.subplots(figsize=(10, 5))
        sns.boxplot(
            data=df_plot.dropna(), x="weekday", y=self.target, hue="season", ax=ax, linewidth=1
        )
        ax.set_title("Effetti settimanali-stagionali")
        ax.set_xlabel("Giorno della settimana")
        ax.set_ylabel(f"Numero {self.use_case}")
        ax.legend(bbox_to_anchor=(1, 1))
        return fig
    
    @log_exception(logger)
    def analyzingTimeSeries(self, aggregation_level: str,) -> dict:
        
        """Perform an analysis of time series (plot,stationarity test,autocorrelation and additive decomposition)

        Parameters
        ----------

        Returns
        -------
            dict
            dict tath contain result of analysis (stationarity, trend and seasonality)
        """

        warnings.filterwarnings("ignore")

        visualizations = {}
        logger.info(
            msg="START elaborate time series plot and time series mirror plot",
            important=True,
        )
        plot = self.plotTimeSeries()
        visualizations["serie_temporale"] = [*plot]
        logger.info(
            msg="DONE elaborate time series plot and time series mirror plot",
            important=True,
        )

        logger.info(msg="START distribution plot", important=True)
        distribution = self.distribution_plot()
        visualizations["serie_temporale"].append(distribution)
        logger.info(msg="DONE distribution plot", important=True)

        if aggregation_level == "D":
            logger.info(
                msg="START elaborate plot for day of week and for mount of year",
                important=True,
            )
            plotDayMonth = self.plot_day_month()
            visualizations["aggregazioni"] = [*plotDayMonth]
            logger.info(
                msg="DONE elaborate plot for day of week and for mount of year",
                important=True,
            )

            logger.info(msg="START weekly-seasonal plot", important=True)
            weekly_seasonal_effects = self.boxplots_seasons()
            visualizations["aggregazioni"].append(weekly_seasonal_effects)
            logger.info(msg="DONE weekly-seasonal plot", important=True)

        if len(self.df) > 5:
            logger.info(msg="START autocorrelation plot", important=True)
            autocorr = self.autocorrelation(n_lags=min(len(self.df.dropna()[self.target]) // 3 - 1, 365))
            visualizations["autocorrelazione"] = [*autocorr]
            logger.info(msg="DONE autocorrelation plot", important=True)

        try:
            logger.info(msg="START stationarity test", important=True)
            stationar = self.stationarity()
            visualizations["test_stazionarietà"] = stationar
            logger.info(msg="DONE stationarity test", important=True)
        except:
            pass

        try:
            logger.info(msg="START additive decomposition", important=True)
            additive = self.additive_decomposition()
            visualizations["decomposizione_additiva"] = [additive[2]]
            logger.info(msg="DONE additive decomposition", important=True)
            return {
                "trend": additive[0],
                "seasonality": additive[1],
                "visualizations": visualizations,
            }
        except:
            pass

        return {"visualizations": visualizations}



import os
import logging
from logging import DEBUG
from time import localtime, strftime
from test_tc.utility.tele_logger import Logger

def test_logger_initialization():
    logger = Logger()
    assert isinstance(logger.logger, logging.Logger)
    assert isinstance(logger.important_logger, logging.Logger)
    assert logger.name is None
    assert logger.log_path is None

def test_logger_initialize_logger(tmpdir):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.info('TEST')
    # Check if log files are created
    assert os.path.exists(tmpdir)

    log_files = os.listdir(tmpdir)
    assert f"test_session_{strftime('%Y%m%d', localtime())}.log" in log_files
    assert f"test_session_verbose_{strftime('%Y%m%d', localtime())}.log" in log_files

def test_logger_info_important(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.info("This is an info message", important = True)

    assert len(caplog.records) == 2
    assert caplog.records[0].message == "This is an info message"
    assert caplog.records[1].message == "This is an info message"

def test_logger_info(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.info("This is an info message")

    assert len(caplog.records) == 1
    assert caplog.records[0].message == "This is an info message"

def test_logger_error(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.error("This is an error message")

    assert len(caplog.records) == 2
    assert caplog.records[0].message == "This is an error message"
    assert caplog.records[1].message == "This is an error message"

def test_logger_exception(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    try:
        1 / 0
    except ZeroDivisionError as e:
        logger.exception("An exception occurred")

    assert len(caplog.records) == 2
    assert caplog.records[0].message == "An exception occurred"
    assert caplog.records[1].message == "An exception occurred"

def test_logger_warning(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.warning("This is a warning message")

    assert len(caplog.records) == 1
    assert caplog.records[0].message == "This is a warning message"

def test_logger_debug_not_init(tmpdir, caplog):
    logger = Logger()
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.debug("This is a debug message")

    assert len(caplog.records) == 0

def test_logger_debug_init(tmpdir, caplog):
    logger = Logger(level=DEBUG)
    logger.initialize_logger(log_path=tmpdir, name="test_session")

    logger.debug("This is a debug message")

    assert len(caplog.records) == 1
    assert caplog.records[0].message == "This is a debug message"



import pandas as pd
from test_tc.datapreparation.uc_preproc import TCPreprocess, TVPreprocess, TAPreprocess, TMPreprocess

def sample_data():
    # Sample data for testing
    data = {
        "column_a": [1, 2, 3],
        "column_b": [4, 5, 6],
        "column_c": [7, 8, 9],
        'target_col': ['a','a','b']
    }
    return pd.DataFrame(data)

def sample_dataTV():
    # Sample data for testing
    data = {
        "column_a": [1, 2, 3],
        "column_b": [4, 5, 6],
        "column_c": ['7', '8', '9'],
        'target_col': ['a','a','b']
    }
    return pd.DataFrame(data)

def test_TCPreprocess():

    hierarchy = {
        "level_1": "column_a",
        "level_2": "column_b",
        "level_3": "column_c"
    }
    conversion = {
        "level_1": None,
        "level_2": None,
        "level_3": "{7: 'seven', 8: 'eight', 9: 'nine'}"
    }

    target_col = "target_col"

    # Sample data for testing
    df = sample_data()
    preprocess = TCPreprocess(hierarchy, conversion, target_col)

    # Test fit method
    preprocess_fit = preprocess.fit()

    # fit() should return the same instance of the TCPreprocess object
    assert preprocess_fit is preprocess

    # Test transform method
    transformed_data = preprocess.transform(df)

    # Check if the columns are correctly mapped from the hierarchy and conversion dictionaries
    for value in transformed_data['column_c']:
        assert value in ['seven', 'eight', 'nine']

    # Check if the target col has been filtered
    assert transformed_data.shape[0] == 2
   
def test_TVPreprocess():

    hierarchy = {
        "level_1": "column_a",
        "level_2": "column_b",
        "level_3": "column_c"
    }
    conversion = {
        "level_1": None,
        "level_2": None,
        "level_3": "{7: 'seven', 8: 'eight', 9: 'nine'}"
    }

    # Sample data for testing
    df = sample_dataTV()
    preprocess = TVPreprocess(hierarchy, conversion)

    # Test fit method
    preprocess_fit = preprocess.fit()

    # fit() should return the same instance of the TVPreprocess object
    assert preprocess_fit is preprocess

    # Test transform method
    transformed_data = preprocess.transform(df)

    # Check if the columns are correctly mapped from the hierarchy and conversion dictionaries
    for value in transformed_data['column_c']:
        assert value in ['seven', 'eight', 'nine']

    # Check if the target col has NOT been filtered
    assert transformed_data.shape[0] == 3

def test_TMPreprocess():
    hierarchy = {
        "level_1": "column_a",
        "level_2": "column_b",
        "level_3": "column_c"
    }
    conversion = {
        "level_1": None,
        "level_2": None,
        "level_3": "{7: 'seven', 8: 'eight', 9: 'nine'}"
    }

    # Sample data for testing
    df = sample_data()

    preprocess = TMPreprocess(hierarchy, conversion)

    # Test fit method
    preprocess_fit = preprocess.fit()

    # fit() should return the same instance of the TMPreprocess object
    assert preprocess_fit is preprocess

    # Test transform method
    transformed_data = preprocess.transform(df)

   # Check if the columns are correctly mapped from the hierarchy and conversion dictionaries
    for value in transformed_data['column_c']:
        assert value in ['seven', 'eight', 'nine']

    # Check if the target col has been filtered
    assert transformed_data.shape[0] == 3

def test_TAPreprocess():
    hierarchy = {
        "level_1": "column_a",
        "level_2": "column_b",
        "level_3": "column_c"
    }
    conversion = {
        "level_1": None,
        "level_2": None,
        "level_3": "{7: 'seven', 8: 'eight', 9: 'nine'}"
    }

    # Sample data for testing
    df = sample_data()
    preprocess = TAPreprocess(hierarchy, conversion)

    # Test fit method
    preprocess_fit = preprocess.fit()

    # fit() should return the same instance of the TAPreprocess object
    assert preprocess_fit is preprocess

    # Test transform method
    transformed_data = preprocess.transform(df)

    # Check if the columns are correctly mapped from the hierarchy and conversion dictionaries
    for value in transformed_data['column_c']:
        assert value in ['seven', 'eight', 'nine']

    # Check if the target col has been filtered
    assert transformed_data.shape[0] == 3


import pytest
from test_tc.utility.configuration_utils import *
from test_tc.utility.exceptions import InputTypeError
import re

project_path = os.path.dirname(os.path.abspath(__file__))
config_path = os.path.join(project_path, 'config')

# test get_configuration ##
def test_get_configuration():
    config_file = "unittest_path.toml"  
    # Test reading a valid configuration file
    expected_config =     {'logging': {'logging_folder': 'logs'},
                           'data_paths': {'data_folder': 'tests/data', 'experiment_folder': 'tests/experiments/mlruns', 'predict_save_folder': 'data/output/output_predictions', 'predict_real_folder': 'tests/data/models'}, 
                           'data_file': {'data_input_filename': 'unittest_teleconsulto.parquet'}}
    
    config = get_configuration(config_path, config_file)
    assert config == expected_config

    # Test reading an invalid configuration file (file not found)
    non_existent_config_file = "non_existent_config.toml"
    with pytest.raises(Exception) as exc_info:
        get_configuration(config_path, non_existent_config_file)
    assert str(exc_info.value) == f"File {non_existent_config_file} not found in {config_path} please double check that the filename is correct."

## test config_reading ##
def test_config_reading():
    # Prepare test data
    config_dict = {
        'config_path': config_path,  # Replace this with the actual path to your test configuration files
        'config_file': "unittest_prophet.toml"  # Replace this with the actual config file name
    }

    config_dict_parsing = {
        'config_path': config_path,  # Replace this with the actual path to your test configuration files
        'config_file': "unittest_prophet_parsing.toml"  # Replace this with the actual config file name for parsing
    }

    # Test reading two valid configuration files
    config_file, config_file_parsing = config_reading(config_dict, config_dict_parsing)

    assert isinstance(config_file, dict)
    assert isinstance(config_file_parsing, dict)

## test config_key_parsing ##
def test_config_key_parsing():
    config_file = {"key1": "value1", "key2": 42, "key3": [1, 2, 3]}

    config_file_parsing = {"key1": ['value1'], "key2": [42], "key3": [1, 2, 3]}

    # Test for matching keys (should not raise an exception)
    config_key_parsing(config_file, config_file_parsing)

    # Test for non-matching keys (should raise an exception)
    config_file_invalid = {"key1": "value1", "key4": "value4",}

    with pytest.raises(InputTypeError):
        config_key_parsing(config_file_invalid, config_file_parsing)

## test check_usecases ##
def test_check_usecases():
    # Prepare test data
    valid_usecase = "usecase1"
    possible_usecases = ["usecase1", "usecase2", "usecase3"]

    # Test for a valid usecase (should not raise an exception)
    check_usecases(valid_usecase, possible_usecases)

    # Test for an invalid usecase (should raise an exception)
    invalid_usecase = "invalid_usecase"

    with pytest.raises(InputTypeError):
        check_usecases(invalid_usecase, possible_usecases)

## test check_model ##
def test_check_model():
    # Prepare test data
    valid_model_name = "model1"
    valid_reference_model_name = "model1"
    valid_apply_hierarchical = True
    valid_reference_apply_hierarchical = [True, False]

    # Test for valid model and apply_hierarchical (should not raise an exception)
    check_model(valid_model_name, valid_reference_model_name, valid_apply_hierarchical, valid_reference_apply_hierarchical)

    # Test for invalid model_name (should raise an exception)
    invalid_model_name = "invalid_model"

    with pytest.raises(InputTypeError):
        check_model(invalid_model_name, valid_reference_model_name, valid_apply_hierarchical, valid_reference_apply_hierarchical)

    # Test for inconsistent model_name and reference_model_name (should raise an exception)
    inconsistent_model_name = "model2"

    with pytest.raises(InputTypeError, match=f"model_name {inconsistent_model_name} MUST be equal to reference_model_name {valid_reference_model_name}"):
        check_model(inconsistent_model_name, valid_reference_model_name, valid_apply_hierarchical, valid_reference_apply_hierarchical)

    # Test for invalid apply_hierarchical (should raise an exception)
    invalid_apply_hierarchical = "not_a_bool"

    with pytest.raises(InputTypeError):
        check_model(valid_model_name, valid_reference_model_name, invalid_apply_hierarchical, valid_reference_apply_hierarchical)

## test check_hyperparameters ##
def test_check_hyperparameters():
    # Prepare test data
    valid_hyperparameters = ["param1", "param2", "param3"]
    valid_possible_hyperparameters = ["param1", "param2", "param3", "param4"]

    # Test for valid hyperparameters (should not raise an exception)
    check_hyperparameters(valid_hyperparameters, valid_possible_hyperparameters)

    # Test for invalid hyperparameter type (should raise an exception)
    invalid_hyperparameters = "not_a_list"

    with pytest.raises(Exception):
        check_hyperparameters(invalid_hyperparameters, valid_possible_hyperparameters)

    # Test for invalid type hyperparameter (should raise an exception)
    invalid_hyperparameter = 5

    with pytest.raises(InputTypeError):
        check_hyperparameters([invalid_hyperparameter], valid_possible_hyperparameters)

## test check_hyperparameters_value ##
def test_check_hyperparameters_value():
    # Prepare test data with valid hyperparameters
    valid_hyperparameters = {"param1": [1, 2, 3],
                             "param2": {"subparam1": 42, "subparam2": 3.14}}

    # Test for valid hyperparameters (should not raise an exception)
    check_hyperparameters_value(valid_hyperparameters)

    # Prepare test data with invalid hyperparameters
    invalid_hyperparameters = {"param1": "not_a_list_or_dict",
                               "param2": 42,}

    # Test for invalid hyperparameters (should raise an exception)
    with pytest.raises(InputTypeError, match="The hyperparameters to specify must be List or Dict"):
        check_hyperparameters_value(invalid_hyperparameters)

## test check_missing_data_strategy ##
def test_check_missing_data_strategy():
    # Prepare test data with valid missing_data_strategy
    valid_string_strategy = "strategy1"
    valid_dict_strategy = {"interpolation": "linear", "order": 1}

    # Prepare valid lists of possible strategies, dictionary keys, and interpolation methods
    valid_possible_strategies = ["strategy1", "strategy2", "strategy3"]
    valid_possible_dictionary_keys = ["interpolation", "order"]
    valid_possible_interpolation_methods = ["linear", "spline"]

    # Test for valid missing_data_strategy (should not raise an exception)
    check_missing_data_strategy(valid_string_strategy, valid_possible_strategies, 
                                valid_possible_dictionary_keys, valid_possible_interpolation_methods)

    check_missing_data_strategy(valid_dict_strategy, valid_possible_strategies, 
                                valid_possible_dictionary_keys, valid_possible_interpolation_methods)

    # Prepare invalid test data
    invalid_type_strategy = tuple("invalid_strategy",)
    with pytest.raises(TypeError, match="Invalid type for 'missing_data_strategy' in 'configuration_utils.check_missing_data_strategy'. It must be str or int or dict"):
        check_missing_data_strategy(invalid_type_strategy, valid_possible_strategies, 
                                    valid_possible_dictionary_keys, valid_possible_interpolation_methods)
    # Prepare invalid test data
    invalid_int_strategy = 10
    with pytest.raises(TypeError, match="Invalid value for missing_data_strategy. It must be equal to 0"):
        check_missing_data_strategy(invalid_int_strategy, valid_possible_strategies, 
                                    valid_possible_dictionary_keys, valid_possible_interpolation_methods)
    
    # Prepare invalid test data
    invalid_string_strategy = "invalid_strategy"
    with pytest.raises(InputTypeError, match=re.escape(f"test_tc.utility.configuration_utils.check_missing_data_strategy encountered an error: Invalid value for {invalid_string_strategy}. Value not found in {valid_possible_strategies} - ")):
        check_missing_data_strategy(invalid_string_strategy, valid_possible_strategies, 
                                    valid_possible_dictionary_keys, valid_possible_interpolation_methods)

    # Prepare invalid test data
    invalid_dict_strategy1 = {"not_interpolation_or_order": "value1", "order": 3}
    invalid_dict_strategy2 = {"interpolation": "invalid_method", "order": 3}
    invalid_dict_strategy3 = {"interpolation": "linear", "order": '3'}
    
    with pytest.raises(InputTypeError, match=re.escape(f"test_tc.utility.configuration_utils.check_missing_data_strategy encountered an error: Invalid value for not_interpolation_or_order. Value not found in {valid_possible_dictionary_keys} - ")):
        check_missing_data_strategy(invalid_dict_strategy1, valid_possible_strategies, 
                                    valid_possible_dictionary_keys, valid_possible_interpolation_methods)

    with pytest.raises(InputTypeError, match=re.escape(f"test_tc.utility.configuration_utils.check_missing_data_strategy encountered an error: Invalid value for invalid_method. Value not found in {valid_possible_interpolation_methods} - ")):
            check_missing_data_strategy(invalid_dict_strategy2, valid_possible_strategies, 
                                        valid_possible_dictionary_keys, valid_possible_interpolation_methods)
            
    with pytest.raises(InputTypeError):
            check_missing_data_strategy(invalid_dict_strategy3, valid_possible_strategies, 
                                        valid_possible_dictionary_keys, valid_possible_interpolation_methods)

## test check_time_granularity ##
def test_check_time_granularity():
    # Prepare test data with valid time_granularity
    valid_time_granularity = "hourly"
    valid_possible_time_granularities = ["hourly", "daily", "weekly", "monthly"]

    # Test for valid time_granularity (should not raise an exception)
    check_time_granularity(valid_time_granularity, valid_possible_time_granularities)

    invalid_possible_time_granularities = "not_a_list"

    # Test for invalid time_granularity type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_time_granularity(123.45, valid_possible_time_granularities)

    # Test for invalid possible_time_granularities type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_time_granularity(valid_time_granularity, invalid_possible_time_granularities)

## test check_preprocessing ##
def test_check_preprocessing():
    # Prepare test data with valid target_col and date_col
    valid_target_col = "target_column"
    valid_date_col = "date_column"

    # Test for valid target_col and date_col (should not raise an exception)
    check_preprocessing(valid_target_col, valid_date_col)

    # Prepare invalid test data
    invalid_target_col = 123.45
    invalid_date_col = ["date_column"]

    # Test for invalid target_col type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_preprocessing(invalid_target_col, valid_date_col)

    # Test for invalid date_col type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_preprocessing(valid_target_col, invalid_date_col)

## test check_split_forecasting ##
def test_check_split_forecasting():
    # Prepare test data with valid input
    valid_time_granularity = "D"
    valid_validation_obs = 20
    valid_test_obs = 10
    valid_num_forecast_periods = 5
    valid_start_date = "2023-07-20"

    # Test for valid input (should not raise an exception)
    check_split_forecasting(valid_time_granularity, valid_validation_obs, valid_test_obs,
                            valid_num_forecast_periods, valid_start_date)

    # Prepare invalid test data
    invalid_time_granularity = 123
    invalid_validation_obs = 0
    invalid_num_forecast_periods = 0

    # Test for invalid time_granularity type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_split_forecasting(invalid_time_granularity, valid_validation_obs, valid_test_obs,
                                valid_num_forecast_periods, valid_start_date)

    # Test for invalid validation_obs value (should raise an exception)
    with pytest.raises(InputTypeError):
        check_split_forecasting(valid_time_granularity, invalid_validation_obs, valid_test_obs,
                                valid_num_forecast_periods, valid_start_date)

    # Test for invalid num_forecast_periods value (should raise an exception)
    with pytest.raises(InputTypeError):
        check_split_forecasting(valid_time_granularity, valid_validation_obs, valid_test_obs,
                                invalid_num_forecast_periods, valid_start_date)

    # Test for invalid start_date type (should raise an exception)
    with pytest.raises(InputTypeError):
        check_split_forecasting(valid_time_granularity, valid_validation_obs, valid_test_obs,
                                valid_num_forecast_periods, 123.45)

## test check_hierarchy ##
def test_check_hierarchy():
    # Prepare test data with valid hierarchy
    valid_hierarchy = {"level1": "category1", "level2": "category2", "level3": "category3"}

    # Test for valid hierarchy (should not raise an exception)
    check_hierarchy(valid_hierarchy)

    # Prepare invalid test data
    invalid_hierarchy = {"level1": 123, "level2": "category2", "level3": "category3"}

    # Test for invalid hierarchy (should raise an exception)
    with pytest.raises(InputTypeError):
        check_hierarchy(invalid_hierarchy)

## test check_conversion ##
def test_check_conversion():
    # Prepare test data with valid conversion
    valid_conversion = {"value1": "converted1", "value2": "converted2", "value3": "converted3"}

    # Test for valid conversion (should not raise an exception)
    check_conversion(valid_conversion)

    # Prepare invalid test data
    invalid_conversion = {"value1": 123, "value2": "converted2", "value3": "converted3"}

    # Test for invalid conversion (should raise an exception)
    with pytest.raises(InputTypeError):
        check_conversion(invalid_conversion)

## test config_parsing ##
# The function config_parsing is an ensamble of the functions above
# It's test will consider only its good execution for the config files: unittest_prophet_parsing.toml, unittest_prophet.toml since it's tested already in its sunfunctions
def test_config_parsing():
    config_filename = 'unittest_prophet.toml'
    config_filename_parsing = 'unittest_prophet_parsing.toml'
    config_parsing(config_path, config_filename, config_filename_parsing)



import pytest
import pandas as pd
import numpy as np

from test_tc.analytics.evaluationMetrics import EvaluateModel
from test_tc.utility.exceptions import InputTypeError, ColumnNotFound

def test_evaluate_model_init():
    # Correct input data
    date_true = "timestamp_true"
    date_pred = "timestamp_pred"
    y_true = "target_true"
    y_pred = "target_pred"
    upper_95 = "upper_ci_95"
    lower_95 = "lower_ci_95"

    # Check if the class initializes correctly with correct input data
    model = EvaluateModel(date_true, date_pred, y_true, y_pred, upper_95, lower_95)
    assert model.date_true == date_true
    assert model.date_pred == date_pred
    assert model.y_true == y_true
    assert model.y_pred == y_pred
    assert model.upper_95 == upper_95
    assert model.lower_95 == lower_95

    # Incorrect input data - Using integers instead of strings
    with pytest.raises(InputTypeError):
        EvaluateModel(2023, "timestamp_pred", "target_true", "target_pred", "upper_ci_95", "lower_ci_95")

    # Incorrect input data - Using booleans instead of strings
    with pytest.raises(InputTypeError):
        EvaluateModel("timestamp_true", "timestamp_pred", True, "target_pred", "upper_ci_95", "lower_ci_95")


def test_make_evaluation_column_not_found():
    # Create a sample DataFrame for testing
    df_real = pd.DataFrame({
        "timestamp_true": pd.date_range(start='2023-01-01', periods=5),
        "target_true": [1, 2, 3, 4, 5]
    })

    # This DataFrame is mispelling the 'target_pred' column
    df_pred = pd.DataFrame({
        "timestamp_pred": pd.date_range(start='2023-01-01', periods=5)
    })

    model = EvaluateModel("timestamp_true", "timestamp_pred", "target_true", "target_pred", "upper_ci_95", "lower_ci_95")

    # Check if the method raises the ColumnNotFound exception
    with pytest.raises(ColumnNotFound):
        model.make_evaluation(df_real, df_pred, "id_pred")


def test_make_evaluation_correct_input():
    # Create a sample DataFrame for testing
    df_real = pd.DataFrame({
        "timestamp_true": pd.date_range(start='2023-01-01', periods=5),
        "target_true": [1, 2, 3, 4, 5],
        'upper_ci_95' : [3, 4, 5, 6, 7],
        'lower_ci_95' : [0, 1, 2, 3, 4]
    })

    df_pred = pd.DataFrame({
        "timestamp_pred": pd.date_range(start='2023-01-01', periods=5),
        "target_pred": [1, 2, 2.5, 4, 5.5],
        'upper_ci_95' : [3, 4, 5, 6, 7],
        'lower_ci_95' : [0, 1, 2, 3, 4]
    })

    model = EvaluateModel("timestamp_true", "timestamp_pred", "target_true", "target_pred", "upper_ci_95", "lower_ci_95")

    # Call the method and get the evaluation DataFrame
    evaluation_df = model.make_evaluation(df_real, df_pred, "id_pred", verbose=False)

    # Check if the evaluation DataFrame has the correct columns
    assert all(col in evaluation_df.columns for col in ['Id_Pred', 'MAE', 'MAPE', 'RMSE', 'MSE', 'R2', 'Percentage Coverage'])

    # Check if the 'Id_Pred' column contains the correct value
    assert evaluation_df['Id_Pred'][0] == 'id_pred'

    # Check if the evaluation metrics are calculated correctly (assuming you know the expected results)
    assert np.isclose(evaluation_df['MAE'][0], 0.2, atol=1e-2)
    assert np.isclose(evaluation_df['MAPE'][0], 0.05, atol=1e-2)
    assert np.isclose(evaluation_df['RMSE'][0], 0.32, atol=1e-2)
    assert np.isclose(evaluation_df['MSE'][0], 0.1, atol=1e-2)
    assert np.isclose(evaluation_df['R2'][0], 0.95, atol=1e-2)

    print(evaluation_df['Percentage Coverage'][0])

    # Assuming you have the expected percentage coverage value for this specific example
    # Replace 'expected_percentage_coverage' with the actual expected value for your data
    expected_percentage_coverage = 100  # For example, 100%
    assert np.isclose(evaluation_df['Percentage Coverage'][0], expected_percentage_coverage, atol=1e-2)









